#include <OSWindow.h>

#include "Win32GraphicsContext.h"
#include "Win32WindowKeyboardInput.h"
#include "Win32WindowMouseInput.h"
#include "ThreadPrimitives.h"

#include <boost/function.hpp>
#include <boost/bind.hpp>
#include <iostream>
#include <windows.h>

static LRESULT CALLBACK WndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);

struct WinClass
{
	const char* className;

	WinClass()
	: className("TopWin32WindowClass")
	{
		WNDCLASSEX wc;
		wc.cbSize = sizeof(wc);
		wc.style = 0;
		wc.lpfnWndProc = WndProc;
		wc.cbClsExtra = 0;
		wc.cbWndExtra = sizeof(Win32Window_Pimpl*) + 8;
		wc.hInstance = GetModuleHandle(NULL);
		wc.hIcon = LoadIcon(NULL, IDI_APPLICATION);
		wc.hCursor = NULL;
		wc.hbrBackground = (HBRUSH)(COLOR_WINDOW+1);
		wc.lpszMenuName = NULL;
		wc.lpszClassName = className;
		wc.hIconSm = LoadIcon(NULL, IDI_APPLICATION);

		bool result = RegisterClassEx(&wc);
		ASSERT(result);
	}
};
static WinClass s_theClass;

class Win32Window_Pimpl : public OSWindow::Pimpl
{
	Thread* m_winThread;

	Mutex m_monitorLock;
	Condition m_setupDoneCond;

	GenericWindow::MousePosition m_lastPoints;

	HINSTANCE m_hInstance;
	HWND m_hwnd;

	Win32GraphicsContext* m_graphicsContext;
	Win32WindowKeyboardInput m_keyboardInput;
	Win32WindowMouseInput* m_mouseInput;

	bool m_waitingToQuit;
	Condition m_quitCondition;

	void MarkToQuit()
	{
		Mutex::ScopedLock lock(m_monitorLock);
	
		m_waitingToQuit = true;
		m_quitCondition.SignalAll();
	}


	void Run()
	{
		RECT windowRect;
		windowRect.left = 0;
		windowRect.top = 0;
		windowRect.right = m_width;
		windowRect.bottom = m_height;
		AdjustWindowRectEx(&windowRect, m_style, FALSE, 0);

		m_hwnd = CreateWindowEx(
			m_windowedExtendedStyle,
			s_theClass.className, 
			m_title.c_str(),
			m_style,
			CW_USEDEFAULT, CW_USEDEFAULT, 
			windowRect.right - windowRect.left, windowRect.bottom - windowRect.top,
			NULL, NULL, m_hInstance, NULL);

		ASSERT(m_hwnd != NULL);

		SetWindowLongPtr(m_hwnd, 0, (LONG)this);
			
		SetCursor(LoadCursor(NULL, IDC_ARROW));

		ShowWindow(m_hwnd, SW_SHOW);

		m_mouseInput = new Win32WindowMouseInput(m_hwnd);

		m_graphicsContext = new Win32GraphicsContext(m_hwnd);

		{
			Mutex::ScopedLock lock(m_monitorLock);
			m_setupDoneCond.SignalAll();
		}

		MSG msg;
		while(GetMessage(&msg, NULL, 0, 0) > 0)
		{
			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}

		MarkToQuit();

		delete m_graphicsContext;
		m_graphicsContext = 0;
		
		delete m_mouseInput;
	}


	enum CustomEvents
	{
		WM_APP_SHOWHIDECURSOR = WM_APP,

	};

	const DWORD m_style;
	const DWORD m_windowedExtendedStyle;
	const DWORD m_fullscreenExtendedStyle;

	int m_width;
	int m_height;
	std::string m_title;

	int m_windowedWidth;
	int m_windowedHeight;

	OSWindow::Mode m_mode;

	boost::function<void ()> m_onWantQuit; // Called when the window wants to close itself

public:
	Win32Window_Pimpl(const boost::function<void ()>& onWantQuit, const std::string& title, int width, int height)
	: m_lastPoints(0,0)
	, m_graphicsContext(0)
	, m_mouseInput(0)
	, m_width(width)
	, m_height(height)
	, m_title(title)
	, m_windowedWidth(m_width)
	, m_windowedHeight(m_height)
	, m_mode(OSWindow::M_WINDOWED)
	, m_style(WS_POPUPWINDOW | WS_CAPTION)
	, m_windowedExtendedStyle(0)
	, m_fullscreenExtendedStyle(WS_EX_APPWINDOW | WS_EX_TOPMOST)
	, m_waitingToQuit(false)
	, m_onWantQuit(onWantQuit)
	{
		ASSERT(!m_onWantQuit.empty());

		Mutex::ScopedLock lock(m_monitorLock);

		m_hInstance = GetModuleHandle(NULL);
		m_winThread = new Thread(boost::bind(&Win32Window_Pimpl::Run, this));

		m_setupDoneCond.Wait(m_monitorLock);
	}

	virtual ~Win32Window_Pimpl()
	{
		SetWindowMode(OSWindow::M_WINDOWED);

		{ // Wait for our main window event handling thread to be done
			Mutex::ScopedLock lock(m_monitorLock);

			PostMessage(m_hwnd, WM_DESTROY, 0, 0);

			while(!m_waitingToQuit)
			{
				m_quitCondition.Wait(m_monitorLock);
			}
		}

		delete m_winThread;
	}

	LRESULT HandleWin32Message(UINT msg, WPARAM wParam, LPARAM lParam)
	{
		switch(msg)
		{
			case WM_CLOSE: m_onWantQuit(); break;
			case WM_DESTROY: PostQuitMessage(0); break;
			case WM_MOUSEMOVE:
			{
				POINTS points = MAKEPOINTS(lParam);
				m_lastPoints.x = points.x;
				m_lastPoints.y = points.y;
			} break;
			case WM_SYSKEYDOWN:
			case WM_KEYDOWN:
			{
				m_keyboardInput.PostKeyDownEvent(wParam, lParam);
			} break;
			case WM_SYSKEYUP:
			case WM_KEYUP:
			{
				m_keyboardInput.PostKeyUpEvent(wParam, lParam);	
			} break;
			case WM_ACTIVATE:
			{
				if(WA_INACTIVE != wParam && m_mouseInput != 0)
				{
					m_mouseInput->AttemptReaquire();
				}
				else
				{
					SetWindowMode(OSWindow::M_WINDOWED);
				}
			} break;
			case WM_APP_SHOWHIDECURSOR:
			{
				SetCursor((wParam == 0 ? NULL : LoadCursor(NULL, IDC_ARROW)));
			} break;
			default: return DefWindowProc(m_hwnd, msg, wParam, lParam); break;
		}

		return 0;
	}

	virtual Win32GraphicsContext& GetGraphicsContext()
	{
		ASSERT(m_graphicsContext != 0);
		return *m_graphicsContext;
	};
	virtual Win32WindowKeyboardInput& GetKeyboardInput()
	{
		return m_keyboardInput;
	}
	virtual MouseInput& GetMouseInput()
	{
		ASSERT(m_mouseInput != 0);
		return *m_mouseInput;
	}
	virtual GenericWindow::MousePosition GetLastMousePosition()
	{
		return m_lastPoints;
	}

	virtual void SetWindowMode(OSWindow::Mode mode)
	{
		Mutex::ScopedLock lock(m_monitorLock);

		if(mode == OSWindow::M_FULLSCREEN && m_mode != OSWindow::M_FULLSCREEN)
		{
			m_mode = OSWindow::M_FULLSCREEN;

			// Hide the cursor
			PostMessage(m_hwnd, WM_APP_SHOWHIDECURSOR, 0,0);

			// Occupy the desktop dimensions
			HDC dc = GetDC(NULL);
			int width = GetDeviceCaps(dc, HORZRES);
			int height = GetDeviceCaps(dc, VERTRES);
			int bpp = GetDeviceCaps(dc, BITSPIXEL);
			ReleaseDC(NULL, dc);

			m_width = width;
			m_height = height;

			m_graphicsContext->ResizeScreen(m_width, m_height);

			// Move window to hide the border
			RECT windowRect;
			windowRect.left = 0;
			windowRect.top = 0;
			windowRect.right = m_width;
			windowRect.bottom = m_height;
			AdjustWindowRectEx(&windowRect, m_style, FALSE, m_fullscreenExtendedStyle);
			MoveWindow(m_hwnd, windowRect.left, windowRect.top, windowRect.right-windowRect.left, windowRect.bottom-windowRect.top, FALSE);

			SetWindowLong(m_hwnd, GWL_EXSTYLE, m_fullscreenExtendedStyle);

			if(SetForegroundWindow(m_hwnd) == FALSE)
			{
				GoWindowed();
			}
		}
		else if(mode == OSWindow::M_WINDOWED && m_mode != OSWindow::M_WINDOWED)
		{
			GoWindowed();
		}
	}

private:
	void GoWindowed()
	{
		m_width = m_windowedWidth;
		m_height = m_windowedHeight;

		m_graphicsContext->ResizeScreen(m_width, m_height);

		RECT windowRect;
		windowRect.left = 0;
		windowRect.top = 0;
		windowRect.right = m_width;
		windowRect.bottom = m_height;
		AdjustWindowRectEx(&windowRect, m_style, FALSE, m_windowedExtendedStyle);
		MoveWindow(m_hwnd, windowRect.left+50, windowRect.top+50, windowRect.right-windowRect.left, windowRect.bottom-windowRect.top, FALSE);

		SetWindowLong(m_hwnd, GWL_EXSTYLE, m_windowedExtendedStyle);
		PostMessage(m_hwnd, WM_APP_SHOWHIDECURSOR, 1,0);

		m_mode = OSWindow::M_WINDOWED;
	}

public:
	virtual OSWindow::Mode GetWindowMode()
	{
		Mutex::ScopedLock lock(m_monitorLock);

		return m_mode;
	}

	virtual float GetWidth()
	{
		return m_width;
	}

	virtual float GetHeight()
	{
		return m_height;
	}
};

static LRESULT CALLBACK WndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	Win32Window_Pimpl* win = (Win32Window_Pimpl*)GetWindowLongPtr(hwnd, 0);
	if(win == 0)
	{
		return DefWindowProc(hwnd, msg, wParam, lParam);
	}

	ASSERT(win != 0);
	return win->HandleWin32Message(msg, wParam, lParam);
}


OSWindow::OSWindow(const std::string& winTitle, int width, int height)
{
	m_pimpl = new Win32Window_Pimpl(boost::bind(&OSWindow::FireQuitEvent, this), winTitle, width, height);
}

OSWindow::~OSWindow()
{
	delete m_pimpl;
}
