#ifndef TOPALGEBRA_H_
#define TOPALGEBRA_H_

#include "Utils.h"
#include <math.h>
#include <ostream>
#include <float.h>

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

#ifndef M_SQRT1_2
#define M_SQRT1_2 0.7071067811865475244
#endif

#include <xmmintrin.h>

template <int M, int N, bool ALIGNED=true>
class Matrix;

// This class represents a vector of floats, of N elements
enum VecType
{
	VT_INSTANCE,
	VT_REFERENCE,
};
template <int N, VecType T, bool A>
struct Elements
{};

template <int N>
struct Elements<N, VT_INSTANCE, true>
{
	ALIGNED(16) float elems[N];
	
	Elements()
	{};

	template<VecType OTHER_TYPE, bool OTHER_A>
	Elements(const Elements<N,OTHER_TYPE,OTHER_A>& host)
	{
		for(int i=0; i<N; ++i) elems[i] = host.elems[i];
	}
	Elements(const float* array)
	{
		for(int i=0; i<N; ++i) elems[i] = array[i];
	}
};
template <int N>
struct Elements<N, VT_INSTANCE, false>
{
	float elems[N];
	
	Elements()
	{};

	template<VecType OTHER_TYPE, bool OTHER_A>
	Elements(const Elements<N,OTHER_TYPE,OTHER_A>& host)
	{
		for(int i=0; i<N; ++i) elems[i] = host.elems[i];
	}
	Elements(const float* array)
	{
		for(int i=0; i<N; ++i) elems[i] = array[i];
	}
};
template <int N>
struct Elements<N, VT_REFERENCE, true>
{
	const float* elems;

	template<VecType OTHER_TYPE>
	Elements(const Elements<N,OTHER_TYPE,true>& host)
	: elems(host.elems)
	{};
	Elements(const float* array)
	: elems(array)
	{};

	private:
	Elements();
};
template <int N>
struct Elements<N, VT_REFERENCE, false>
{
	const float* elems;

	template<VecType OTHER_TYPE, bool OTHER_A>
	Elements(const Elements<N,OTHER_TYPE,OTHER_A>& host)
	: elems(host.elems)
	{};
	Elements(const float* array)
	: elems(array)
	{};

	private:
	Elements();
};

template <int N, VecType TYPE, bool A>
class VecParent
{
	public:
	Elements<N, TYPE, A> m_elems;

	protected:
	friend class VecParent<N,VT_REFERENCE,false>;
	friend class VecParent<N,VT_INSTANCE,false>;
	friend class VecParent<N,VT_REFERENCE,true>;
	friend class VecParent<N,VT_INSTANCE,true>;

	VecParent()
	{};

	VecParent(const float elems[N])
	: m_elems(elems)
	{};

	template <VecType OTHER_TYPE, bool OTHER_A>
	VecParent(const VecParent<N, OTHER_TYPE, OTHER_A>& other)
	: m_elems(other.m_elems)
	{};
};

template <int N, VecType TYPE=VT_INSTANCE, bool A=true>
class Vector : public VecParent<N,TYPE,A>
{
public:
	Vector()
	{
		for(int i=0; i<N; ++i) VecParent<N,TYPE,A>::m_elems.elems[i] = 0;
	}

	Vector(const float elems[N])
		: VecParent<N,TYPE,A>(elems)
	{}

	template <VecType OTHER_TYPE, bool OTHER_A>
	Vector(const Vector<N, OTHER_TYPE, OTHER_A>& other)
		: VecParent<N,TYPE,A>(other)
	{}


	Vector(float x, float y); // Vector<2> only
	Vector(float x, float y, float z); // Vector<3> only
	Vector(float x, float y, float z, float w); // Vector<4> only

	const float* GetElements() const {return VecParent<N,TYPE,A>::m_elems.elems;}

	void Homogenize(); // Only defined for Vector<4>
	Vector<N> Homogenized() const // Returns the homogenized vector instead of modifying in place
	{
		Vector<N> ret;

		float scale = 1/VecParent<N,TYPE,A>::m_elems.elems[N-1];
		for(int i=0; i<N-1; ++i)
		{
			ret[i] = VecParent<N,TYPE,A>::m_elems.elems[i] * scale;
		}
		ret.m_elems.elems[N-1] = 1;

		return ret;
	}

	const float& operator[](int index) const {ASSERT(index >= 0); ASSERT(index < N); return VecParent<N,TYPE,A>::m_elems.elems[index];}
	float& operator[](int index) {ASSERT(index >= 0); ASSERT(index < N); return VecParent<N,TYPE,A>::m_elems.elems[index];}

	const Vector<N-1, VT_REFERENCE, A> Trim() const
	{
		return Vector<N-1, VT_REFERENCE, A>(VecParent<N,TYPE,A>::m_elems.elems);
	}

	template<VecType OTHER_TYPE, bool OTHER_A>
	void operator+=(const Vector<N, OTHER_TYPE, OTHER_A>& other)
	{
		for(int i=0; i<N; ++i)
		{
			VecParent<N,TYPE,A>::m_elems.elems[i] += other.m_elems.elems[i];
		}
	}

	template<VecType OTHER_TYPE, bool OTHER_A>
	Vector<N> operator-(const Vector<N,OTHER_TYPE,OTHER_A>& other) const
	{
		Vector<N> ret;
		for(int i=0; i<N; ++i)
		{
			ret.m_elems.elems[i] = VecParent<N,TYPE,A>::m_elems.elems[i] - other.m_elems.elems[i];
		}
		return ret;
	}

	template<VecType OTHER_TYPE, bool OTHER_A>
	void operator-=(const Vector<N, OTHER_TYPE, OTHER_A>& other)
	{
		for(int i=0; i<N; ++i)
		{
			VecParent<N,TYPE,A>::m_elems.elems[i] -= other.m_elems.elems[i];
		}
	}

	Vector<N> operator-() const
	{
		Vector<N> ret;
		for(int i=0; i<N; ++i)
		{
			ret.m_elems.elems[i] = -VecParent<N,TYPE,A>::m_elems.elems[i];
		}
		return ret;
	}

	Vector<N> operator*(float scale) const
	{
		Vector<N> ret;
		for(int i=0; i<N; ++i)
		{
			ret.m_elems.elems[i] = VecParent<N,TYPE,A>::m_elems.elems[i] * scale;
		}
		return ret;
	}
	void operator*=(float scale)
	{
		for(int i=0; i<N; ++i) VecParent<N,TYPE,A>::m_elems.elems[i] *= scale;
	}

	Vector<N> operator/(float divisor) const
	{
		ASSERT(divisor != 0);

		float scale = 1/divisor;
		return operator*(scale);
	}
	void operator/=(float divisor)
	{
		ASSERT(divisor != 0);

		float scale = 1/divisor;
		operator*=(scale);
	}

	void Zero() // Makes the vector a zero vector
	{
		for(int i=0; i<N; ++i) VecParent<N,TYPE,A>::m_elems.elems[i] = 0;
	}

	float LengthSquared() const
	{
		float ret = 0;
		for(int i=0; i<N; ++i)
		{
			ret += VecParent<N,TYPE,A>::m_elems.elems[i] * VecParent<N,TYPE,A>::m_elems.elems[i];
		}
		return ret;
	}

	float Length() const
	{
		return sqrt(LengthSquared());
	}

	void Normalize()
	{
		ASSERT(Length() != 0);
		float scale = 1 / Length();
		for(int i=0; i<N; ++i)
		{
			VecParent<N,TYPE,A>::m_elems.elems[i] *= scale;
		}
	}

	Vector<N> Normalized() const
	{
		float magnitudeRecip = 1/Length();
		
		Vector<N> ret;
		for(int i=0; i<N; ++i)
		{
			ret.m_elems.elems[i] = VecParent<N,TYPE,A>::m_elems.elems[i] * magnitudeRecip;
		}
		return ret;
	}
};

template<int N, VecType LOTHER_TYPE, bool LOTHER_A, VecType ROTHER_TYPE, bool ROTHER_A>
Vector<N> operator+(const Vector<N, LOTHER_TYPE, LOTHER_A>& ls, const Vector<N, ROTHER_TYPE, ROTHER_A>& rs)
{
	Vector<N> ret;

	for(int i=0; i<N; ++i)
	{
		ret.m_elems.elems[i] = ls.m_elems.elems[i] + rs.m_elems.elems[i];
	}
	return ret;
}
template<VecType LOTHER_TYPE, VecType ROTHER_TYPE>
Vector<4> operator+(const Vector<4, LOTHER_TYPE, true>& ls, const Vector<4, ROTHER_TYPE, true>& rs)
{
	Vector<4> ret;

	*((__m128*)ret.m_elems.elems) = _mm_add_ps(*((__m128*)ls.m_elems.elems), *((__m128*)rs.m_elems.elems));

	return ret;
}

template <int N, VecType LOTHER_TYPE, bool LOTHER_A, VecType ROTHER_TYPE, bool ROTHER_A>
float Dot(const Vector<N, LOTHER_TYPE, LOTHER_A>& ls, const Vector<N, ROTHER_TYPE, ROTHER_A>& rs)
{
	float ret = 0;
	for(int i=0; i<N; ++i) ret += ls[i] * rs[i];
	return ret;
};

template <>
inline Vector<2, VT_INSTANCE, true>::Vector(float x, float y)
{
	VecParent<2,VT_INSTANCE,true>::m_elems.elems[0] = x;
	VecParent<2,VT_INSTANCE,true>::m_elems.elems[1] = y;
}
template <>
inline Vector<2, VT_INSTANCE, false>::Vector(float x, float y)
{
	VecParent<2,VT_INSTANCE,false>::m_elems.elems[0] = x;
	VecParent<2,VT_INSTANCE,false>::m_elems.elems[1] = y;
}

template <>
inline Vector<3, VT_INSTANCE, true>::Vector(float x, float y, float z)
{
	VecParent<3,VT_INSTANCE,true>::m_elems.elems[0] = x;
	VecParent<3,VT_INSTANCE,true>::m_elems.elems[1] = y;
	VecParent<3,VT_INSTANCE,true>::m_elems.elems[2] = z;
}
template <>
inline Vector<3, VT_INSTANCE, false>::Vector(float x, float y, float z)
{
	VecParent<3,VT_INSTANCE,false>::m_elems.elems[0] = x;
	VecParent<3,VT_INSTANCE,false>::m_elems.elems[1] = y;
	VecParent<3,VT_INSTANCE,false>::m_elems.elems[2] = z;
}

template <>
inline Vector<4, VT_INSTANCE, true>::Vector(float x, float y, float z, float w)
{
	VecParent<4,VT_INSTANCE,true>::m_elems.elems[0] = x;
	VecParent<4,VT_INSTANCE,true>::m_elems.elems[1] = y;
	VecParent<4,VT_INSTANCE,true>::m_elems.elems[2] = z;
	VecParent<4,VT_INSTANCE,true>::m_elems.elems[3] = w;
}
template <>
inline Vector<4, VT_INSTANCE, false>::Vector(float x, float y, float z, float w)
{
	VecParent<4,VT_INSTANCE,false>::m_elems.elems[0] = x;
	VecParent<4,VT_INSTANCE,false>::m_elems.elems[1] = y;
	VecParent<4,VT_INSTANCE,false>::m_elems.elems[2] = z;
	VecParent<4,VT_INSTANCE,false>::m_elems.elems[3] = w;
}

inline float Dot(const Vector<3>& ls, const Vector<3>& rs)
{
	return ls[0]*rs[0] + ls[1]*rs[1] + ls[2]*rs[2];
}

inline Vector<3> Cross(const Vector<3>& ls, const Vector<3>& rs)
{
	return Vector<3>(ls[1]*rs[2] - ls[2]*rs[1], ls[2]*rs[0] - ls[0]*rs[2], ls[0]*rs[1] - ls[1]*rs[0]);
}

template <>
inline void Vector<4,VT_INSTANCE,true>::Homogenize()
{
	float scale = 1/m_elems.elems[3];
	for(int i=0; i<3; ++i)
	{
		m_elems.elems[i] *= scale;
	}
	m_elems.elems[3] = 1;
}
template <>
inline void Vector<4,VT_INSTANCE,false>::Homogenize()
{
	float scale = 1/m_elems.elems[3];
	for(int i=0; i<3; ++i)
	{
		m_elems.elems[i] *= scale;
	}
	m_elems.elems[3] = 1;
}

template<int N, VecType TYPE, bool A>
std::ostream& operator<<(std::ostream& os, const Vector<N, TYPE, A>& vec)
{
	os << "[";
	for(int i=0; i<N-1; ++i)
	{
		os << vec[i] << ", ";
	}
	os << vec[N-1] << "]";

	return os;
}


template <int M, int N, bool ALIGNED>
struct MatElements
{
};

template <int M, int N>
struct MatElements<M, N, true>
{
	ALIGNED(16) float elems[M*N];
};
template <int M, int N>
struct MatElements<M, N, false>
{
	float elems[M*N];
};

template <int M, int N, bool ALIGNED>
class Matrix
{
	MatElements<M,N,ALIGNED> m_elems;

	friend Matrix<4,4> operator*(const Matrix<4,4,true>& ls, const Matrix<4,4,true>& rs);

	public:
	Matrix()
	{
		// Default constructor constructs the identity
		memset(&m_elems, 0, sizeof(m_elems));

		int min = (M < N ? M : N);
		for(int i=0; i<min; ++i)
		{
			m_elems.elems[i*M + i] = 1;
		}
	}

	typedef void (*MatrixMaker)(float[M*N]);
	Matrix(MatrixMaker matFunc) // Construct via a function
	{
		matFunc(m_elems.elems);
	}
	
	static void Identity(float elems[M*N])
	{
		for(int i=0; i<M*N; ++i)
		{
			elems[i] = 0;
		}

		int min = (M < N ? M : N);
		for(int i=0; i<min; ++i)
		{
			elems[i*M + i] = 1;
		}
	}

	static void Zero(float elems[M*N])
	{
		for(int i=0; i<M*N; ++i) elems[i] = 0;
	}

	template<bool OTHER_A>
	Matrix(const Matrix<M,N,OTHER_A>& other)
	{
		operator=(other);
	}

	Matrix(const float elems[M*N])
	{
		memcpy(&m_elems, elems, sizeof(m_elems));
	}

	friend class Matrix<M,N,!ALIGNED>;

	template<bool OTHER_A>
	const Matrix<M,N,ALIGNED>& operator =(const Matrix<M,N,OTHER_A>& other)
	{
		memcpy(&m_elems, &other.m_elems, sizeof(m_elems));
		
		return *this;
	}

	const float* GetElements() const
	{
		return m_elems.elems;
	}

	const Vector<M,VT_REFERENCE,ALIGNED> GetColumn(int col) const
	{
		return Vector<M,VT_REFERENCE,ALIGNED>(&m_elems.elems[col * M]);
	}

	Vector<N> GetRow(int row) const
	{
		Vector<M> ret;

		for(int i=0; i<N; ++i)
		{
			ret[i] = GetVal(row, i);
		}
		return ret;
	}

	template<VecType VEC_TYPE>
	void SetColumn(int colNum, const Vector<M,VEC_TYPE>& col)
	{
		int destStart = colNum * M;
		for(int i=0; i<M; ++i)
		{
			m_elems.elems[destStart + i] = col[i];
		}
	}

	float GetVal(int row, int col) const
	{
		ASSERT(0 <= row && row < M);
		ASSERT(0 <= col && col < N);
		return GetVal(col*M + row);
	}
	float& GetVal(int row, int col)
	{
		ASSERT(0 <= row && row < M);
		ASSERT(0 <= col && col < N);
		return GetVal(col*M + row);
	}

	float GetVal(int index) const
	{
		ASSERT(0 <= index && index < M*N);
		return m_elems.elems[index];
	}
	float& GetVal(int index)
	{
		return m_elems.elems[index];
	}

	Matrix<M,N> operator-() const
	{
		Matrix<M,N> ret;
		for(int i=0; i<M*N; ++i) ret.m_elems.elems[i] = m_elems.elems[i];
		return ret;
	}
	
	Matrix<M,N> operator*(float val) const
	{
		Matrix<M,N> ret;
		for(int i=0; i<M*N; ++i) ret.m_elems.elems[i] = m_elems.elems[i]*val;
		return ret;
	}

	void operator*=(float val)
	{
		for(int i=0; i<M*N; ++i) m_elems.elems[i] *= val;
	}

	Matrix<N,M> GetTranspose() const
	{
		Matrix<N,M> ret;

		int index=0;
		for(int c=0; c<N; ++c)
		{
			for(int r=0; r<M; ++r)
			{
				ret.GetVal(c, r) = GetVal(index);
				++index;
			}
		}

		return ret;
	}

	Matrix<M,N> Inverse() const
	{
		ASSERT(M == N);
		
		Matrix<M,M> leftSide(GetTranspose());
		Matrix<M,M> rightSide(Matrix<M,M>::Identity);

		// Gauss-Jordan
		for(int d=0; d<M; ++d)
		{
			// First, swap rows so that we divide with the largest number
			int biggest=d;
			for(int i=d+1; i<M; ++i)
			{
				if(abs(leftSide.GetColumn(i)[d]) > abs(leftSide.GetColumn(biggest)[d]))
				{
					biggest = i;
				}
			}
			if(biggest != d)
			{
				Vector<M> swap = leftSide.GetColumn(d);
				leftSide.SetColumn(d, leftSide.GetColumn(biggest));
				leftSide.SetColumn(biggest, swap);
				
				swap = rightSide.GetColumn(d);
				rightSide.SetColumn(d, rightSide.GetColumn(biggest));
				rightSide.SetColumn(biggest, swap);
			}

			
			rightSide.SetColumn(d, rightSide.GetColumn(d) / leftSide.GetColumn(d)[d]);
			leftSide.SetColumn(d, leftSide.GetColumn(d) / leftSide.GetColumn(d)[d]);
	
			for(int i=0; i<M; ++i)
			{
				if(i == d) continue;
		
				rightSide.SetColumn(i, rightSide.GetColumn(i) - rightSide.GetColumn(d) * leftSide.GetColumn(i)[d]);
				leftSide.SetColumn(i, leftSide.GetColumn(i) - leftSide.GetColumn(d) * leftSide.GetColumn(i)[d]);
			}
		}

		return rightSide.GetTranspose();
	}

	// Returns a sub-matrix of this matrix, from rows R1 to R2 (inclusive) and columns C1 to C2
	template<int R1, int C1, int R2, int C2>
	Matrix<R2-R1+1, C2-C1+1> SubMatrix() const
	{
		Matrix<R2-R1+1, C2-C1+1> ret;

		for(int c=0; c<C2-C1+1; ++c)
		{
			for(int r=0; r<R2-R1+1; ++r)
			{
				ret.GetVal(r,c) = GetVal(R1+r, C1+c);
			}
		}
		
		return ret;
	}
};

template<int M, int R, int N, bool L_A, bool R_A>
Matrix<M,R> operator*(const Matrix<M,N,L_A>& ls, const Matrix<N,R,R_A>& rs)
{
	Matrix<M,R> ret;

	int outIndex = 0;
	for(int outCol = 0; outCol < R; ++outCol)
	{
		for(int outRow = 0; outRow < M; ++outRow)
		{
			float outVal = 0;
			for(int in = 0; in < N; ++in)
			{
				outVal += ls.GetVal(outRow, in) * rs.GetVal(in, outCol);
			}

			ret.GetVal(outIndex) = outVal;

			++outIndex;
		}
	}

	return ret;
}

inline Matrix<4,4> operator*(const Matrix<4,4,true>& ls, const Matrix<4,4,true>& rs)
{
	// optimized for SSE2
	Matrix<4,4> ret;

	__m128* c1 = (__m128*)&ret.m_elems.elems[0];
	__m128* c2 = (__m128*)&ret.m_elems.elems[4];
	__m128* c3 = (__m128*)&ret.m_elems.elems[8];
	__m128* c4 = (__m128*)&ret.m_elems.elems[12];


	*c1 = _mm_mul_ps(_mm_set_ps1(rs.m_elems.elems[0]), *(__m128*)&ls.m_elems.elems[0]);
	*c1 = _mm_add_ps(_mm_mul_ps(_mm_set_ps1(rs.m_elems.elems[1]), *(__m128*)&ls.m_elems.elems[4]), *c1);
	*c1 = _mm_add_ps(_mm_mul_ps(_mm_set_ps1(rs.m_elems.elems[2]), *(__m128*)&ls.m_elems.elems[8]), *c1);
	*c1 = _mm_add_ps(_mm_mul_ps(_mm_set_ps1(rs.m_elems.elems[3]), *(__m128*)&ls.m_elems.elems[12]), *c1);
	
	*c2 = _mm_mul_ps(_mm_set_ps1(rs.m_elems.elems[4]), *(__m128*)&ls.m_elems.elems[0]);
	*c2 = _mm_add_ps(_mm_mul_ps(_mm_set_ps1(rs.m_elems.elems[5]), *(__m128*)&ls.m_elems.elems[4]), *c2);
	*c2 = _mm_add_ps(_mm_mul_ps(_mm_set_ps1(rs.m_elems.elems[6]), *(__m128*)&ls.m_elems.elems[8]), *c2);
	*c2 = _mm_add_ps(_mm_mul_ps(_mm_set_ps1(rs.m_elems.elems[7]), *(__m128*)&ls.m_elems.elems[12]), *c2);
	
	*c3 = _mm_mul_ps(_mm_set_ps1(rs.m_elems.elems[8]), *(__m128*)&ls.m_elems.elems[0]);
	*c3 = _mm_add_ps(_mm_mul_ps(_mm_set_ps1(rs.m_elems.elems[9]), *(__m128*)&ls.m_elems.elems[4]), *c3);
	*c3 = _mm_add_ps(_mm_mul_ps(_mm_set_ps1(rs.m_elems.elems[10]), *(__m128*)&ls.m_elems.elems[8]), *c3);
	*c3 = _mm_add_ps(_mm_mul_ps(_mm_set_ps1(rs.m_elems.elems[11]), *(__m128*)&ls.m_elems.elems[12]), *c3);

	*c4 = _mm_mul_ps(_mm_set_ps1(rs.m_elems.elems[12]), *(__m128*)&ls.m_elems.elems[0]);
	*c4 = _mm_add_ps(_mm_mul_ps(_mm_set_ps1(rs.m_elems.elems[13]), *(__m128*)&ls.m_elems.elems[4]), *c4);
	*c4 = _mm_add_ps(_mm_mul_ps(_mm_set_ps1(rs.m_elems.elems[14]), *(__m128*)&ls.m_elems.elems[8]), *c4);
	*c4 = _mm_add_ps(_mm_mul_ps(_mm_set_ps1(rs.m_elems.elems[15]), *(__m128*)&ls.m_elems.elems[12]), *c4);

	return ret;
}

template<int M, int N, int R, VecType VEC_TYPE, bool VEC_A>
Vector<R> operator*(const Matrix<M,N>& mat, const Vector<R,VEC_TYPE,VEC_A>& vec)
{
	Vector<R> ret;

	for(int r=0; r<R; ++r)
	{
		float outVal = 0;
		for(int i=0; i<R; ++i)
		{
			outVal += mat.GetVal(r,i) * vec[i];
		}

		ret[r] = outVal;
	}

	return ret;
}


template<int M, int N>
std::ostream& operator <<(std::ostream& os, const Matrix<M,N>& mat)
{
	for(int r=0; r<M; ++r)
	{
		os << "[";
		for(int c=0; c<N; ++c)
		{
			os << "\t" << mat.GetVal(r,c);
		}
		os << "\t" << "]" << std::endl;
	}

	return os;
}

#include <Tuple.h>
#include <complex>

// I bet this could be sweetly templatized by using the adjoint matrix and recursive templates... but fuck that, this works for everything I need.
inline float GetDeterminant(const Matrix<3,3>& a)
{
	return a.GetVal(0,0)*a.GetVal(1,2)*a.GetVal(2,1) + a.GetVal(0,1)*a.GetVal(1,0)*a.GetVal(2,2)
			+ a.GetVal(0,2)*a.GetVal(1,1)*a.GetVal(2,0)
			-a.GetVal(0,0)*a.GetVal(1,1)*a.GetVal(2,2) -a.GetVal(0,1)*a.GetVal(1,2)*a.GetVal(2,0)
			-a.GetVal(0,2)*a.GetVal(1,0)*a.GetVal(2,0);
}

inline Tuple<std::complex<double>, 3> GetCharPoly(const Matrix<3,3>& a) // Returns the characteristic polynomial of a 3x3 matrix
{
	Tuple<std::complex<double>, 3> ret;

	ret[0] = -a.GetVal(2,2) - a.GetVal(1,1) - a.GetVal(0,0);
	ret[1] = a.GetVal(0,0) * a.GetVal(2,2) + a.GetVal(1,1)*a.GetVal(2,2) + a.GetVal(0,0)*a.GetVal(1,1)
			-a.GetVal(1,2)*a.GetVal(2,1) -a.GetVal(0,1)*a.GetVal(1,0) -a.GetVal(0,2)*a.GetVal(2,0);
	ret[2] = a.GetVal(0,0)*a.GetVal(1,2)*a.GetVal(2,1) + a.GetVal(0,1)*a.GetVal(1,0)*a.GetVal(2,2)
			+ a.GetVal(0,2)*a.GetVal(1,1)*a.GetVal(2,0)
			-a.GetVal(0,0)*a.GetVal(1,1)*a.GetVal(2,2) -a.GetVal(0,1)*a.GetVal(1,2)*a.GetVal(2,0)
			-a.GetVal(0,2)*a.GetVal(1,0)*a.GetVal(2,1);

	return ret;
}

// Return value is a Tuple for array efficiency...  really this is a set

// x^3 + a2 * x^2 + a1 * x + a0 = 0
inline Tuple<std::complex<double>, 3> SolveCubic(const std::complex<double>& a2, const std::complex<double>& a1, const std::complex<double>& a0)
{
	const std::complex<double> p = (3.0*a1 - a2*a2) / 3.0;
	const std::complex<double> q = (9.0*a1*a2 - 27.0*a0 - 2.0*a2*a2*a2) / 27.0;

	std::complex<double> sols[3];

	const double EPSILON = 0.00000000000000000000001;
	if(abs(real(p))+abs(imag(p)) < EPSILON) // if p is 0
	{
		sols[0] = pow(q, 1.0/3.0);
		double wabs = abs(sols[0]);
		double warg = arg(sols[0]);

		// Now find the other solutions
		for(int i=1; i<3; ++i)
		{
			sols[i] = std::polar(wabs, warg + i*2.0*M_PI/3.0);
		}
	}
	else
	{
		const std::complex<double> wCubed = q/2.0 + sqrt((q*q)/4.0 + (p*p*p)/27.0);
		sols[0] = pow(wCubed, 1.0/3.0);
		double wabs = abs(sols[0]);
		double warg = arg(sols[0]);

		// Now find the other solutions
		for(int i=1; i<3; ++i)
		{
			sols[i] = std::polar(wabs, warg + i*2.0*M_PI/3.0);
		}

		for(int i=0; i<3; ++i)
		{
			sols[i] = sols[i] - (p / (3.0 * sols[i]));
		}
	}

	for(int i=0; i<3; ++i)
	{
		sols[i] = sols[i] - a2 / 3.0;
	}

	// Now strip the duplicates
	return MakeTuple<std::complex<double> >(sols[0],sols[1],sols[2]);
}

inline Tuple<std::complex<double>, 3> SolveCubic(const std::complex<double>& a3, const std::complex<double>& a2, const std::complex<double>& a1, const std::complex<double>& a0)
{
	return SolveCubic(a2/a3, a1/a3, a0/a3);
}

inline Tuple<std::complex<double>, 3> GetEigenvalues(const Matrix<3,3>& mat)
{
	Tuple<std::complex<double>, 3> charPoly = GetCharPoly(mat);
		
	return SolveCubic(charPoly[0], charPoly[1], charPoly[2]);
}

inline Tuple<std::complex<double>, 3> GetEigenvaluesOfLinear(const Matrix<4,4>& mat)
{
	Matrix<3,3> innerMat;
	innerMat.SetColumn(0, mat.GetColumn(0).Trim());
	innerMat.SetColumn(1, mat.GetColumn(1).Trim());
	innerMat.SetColumn(2, mat.GetColumn(2).Trim());
	
	return GetEigenvalues(innerMat);
}

template <class T>
bool NumbersClose(const T& a, const T& b, const T& epsilon)
{
	T diff = a - b;
	return (diff < 0 ? diff > -epsilon : diff < epsilon);
}



template <bool ALIGNED=true>
class UnitQuaternionBase
{
	friend class UnitQuaternionBase<false>;
	friend class UnitQuaternionBase<true>;

	Vector<3, VT_INSTANCE, ALIGNED> m_v;
	float m_w;

	float Magnitude() const
	{
		return m_v.LengthSquared() + m_w*m_w;
	}

	void AssertSane()
	{
		ASSERT(_finite(m_w));
		ASSERT(!_isnan(m_w));

		for(int i=0; i<3; ++i)
		{
			ASSERT(_finite(m_v[i]));
			ASSERT(!_isnan(m_v[i]));
		}
	}
	void AssertNormalized()
	{
		ASSERTMSG(abs(1.0f - Magnitude()) < 0.0001f, "Must be a unit vector");
	}

public:
	UnitQuaternionBase()
		: m_v(0,0,0)
		, m_w(1.0f)
	{};

	UnitQuaternionBase(float x, float y, float z, float w)
		: m_v(x,y,z)
		, m_w(w)
	{};

	UnitQuaternionBase(const Vector<4>& elems)
		: m_v(elems.Trim())
		, m_w(elems[3])
	{
		AssertSane();
		AssertNormalized();
	}
	UnitQuaternionBase(const Vector<3>& v, float w)
		: m_v(v)
		, m_w(w)
	{
		AssertSane();

		if(abs(1.0f - Magnitude()) >= 0.0001f)
		{
			m_w = sqrt(1.0f - m_v.LengthSquared());
		}
		
		AssertNormalized();
	}
	UnitQuaternionBase(const Vector<3>& v)
		: m_v(v.LengthSquared() > 1.0f ? v.Normalized() : v)
		, m_w(sqrt(1-m_v.LengthSquared()))
	{
		AssertSane();
		AssertNormalized();
	};

	template <bool OTHER_ALIGNED>
	UnitQuaternionBase(const UnitQuaternionBase<OTHER_ALIGNED>& other)
		: m_v(other.m_v)
		, m_w(other.m_w)
	{
	};

	template <bool OTHER_ALIGNED>
	const UnitQuaternionBase<ALIGNED>& operator=(const UnitQuaternionBase<OTHER_ALIGNED>& other)
	{
		m_v = other.m_v;
		m_w = other.m_w;
		return *this;
	}
	
	UnitQuaternionBase<true> Conjugate() const
	{
		return UnitQuaternionBase<true>(-m_v, m_w);
	}

	const Vector<3, VT_INSTANCE, ALIGNED>& GetV() const {return m_v;}
	float GetW() const {return m_w;}

	template <bool OTHER_ALIGNED>
	const UnitQuaternionBase<true> operator*(const UnitQuaternionBase<OTHER_ALIGNED> &rs) const
	{
		return UnitQuaternionBase<true>(Cross(m_v, rs.m_v) + m_v*rs.m_w + rs.m_v*m_w, m_w*rs.m_w - Dot(m_v, rs.m_v));
	}

	const UnitQuaternionBase<true> Inverse() const
	{
		float invNormSquared = 1.0f / (m_v[0]*m_v[0] + m_v[1]*m_v[1] + m_v[2]*m_v[2] + m_w*m_w);
		return UnitQuaternionBase<true>(-m_v[0]*invNormSquared, -m_v[1]*invNormSquared, -m_v[2]*invNormSquared, m_w*invNormSquared);
	}

	Matrix<4,4> ToMatrix() const
	{
		Matrix<4,4> retMatrix;

		double xx = m_v[0]*m_v[0];
		double yy = m_v[1]*m_v[1];
		double zz = m_v[2]*m_v[2];
		double xy = m_v[0]*m_v[1];
		double xz = m_v[0]*m_v[2];
		double yz = m_v[1]*m_v[2];
		double wx = m_w*m_v[0];
		double wy = m_w*m_v[1];
		double wz = m_w*m_v[2];

		retMatrix.GetVal(0,0) = 1 - 2*(yy + zz);
		retMatrix.GetVal(0,1) = 2*(xy - wz);
		retMatrix.GetVal(0,2) = 2*(xz + wy);

		retMatrix.GetVal(1,0) = 2*(xy + wz);
		retMatrix.GetVal(1,1) = 1 - 2*(xx + zz);
		retMatrix.GetVal(1,2) = 2*(yz - wx);

		retMatrix.GetVal(2,0) = 2*(xz - wy);
		retMatrix.GetVal(2,1) = 2*(yz + wx);
		retMatrix.GetVal(2,2) = 1 - 2*(xx + yy);

		retMatrix.GetVal(3,3) = 1;

		return retMatrix;
	}
};

typedef UnitQuaternionBase<true> UnitQuaternion;
typedef UnitQuaternionBase<false> UnalignedUnitQuaternion;

#endif
