#ifndef BULKFILEIO
#define BULKFILEIO

#include <ResourceIdentifier.h>

#include <ThreadPrimitives.h>

#include <map>

namespace BulkFileIO
{
	class FileMemAllocator
	{
		const long m_maxMem;
		long m_curMem;

		Mutex m_monitorMutex;
		Condition m_freedCond;

		std::map<void*, long> m_memSizes;

		// Platform dependant implementation
		void* PlatAllocate(long numBytes);
		void PlatFree(void* mem);

	public:
		FileMemAllocator(const long maxMem)
			: m_maxMem(maxMem)
			, m_curMem(0)
		{};
		~FileMemAllocator()
		{
			ASSERT(m_curMem == 0);
		}
		void* Allocate(long numBytes)
		{
			Mutex::ScopedLock lock(m_monitorMutex);
			ASSERT(numBytes <= m_maxMem);

			while(m_curMem + numBytes > m_maxMem)
			{
				m_freedCond.Wait(m_monitorMutex);
			}

			m_curMem += numBytes;

			void* alloc = PlatAllocate(numBytes);

			m_memSizes[alloc] = numBytes;

			return alloc;
		}

		void Free(void* mem)
		{
			Mutex::ScopedLock lock(m_monitorMutex);

			PlatFree(mem);

			std::map<void*, long>::iterator found = m_memSizes.find(mem);
			ASSERT(found != m_memSizes.end());

			m_curMem -= found->second;

			m_memSizes.erase(found);

			m_freedCond.SignalAll();
		}
	};

	struct LoadResults
	{
		long dataSize;
		void* data;
	};

	LoadResults LoadFileToMemory(const ResourceIdentifier& resId, FileMemAllocator& alloc);
};

#endif
