#ifndef MYEVENTS_H_
#define MYEVENTS_H_

#include <Utils.h>

#include <boost/function.hpp>

#include <memory>
#include <list>


template <class FunctionSignature>
class Event;

namespace EventDetails
{
	// Clients create EventConnections with the Events (or Event Ports) to signify their desire to
	// be notified when the event is fired
	template <class FunctionSignature>
	class Connection
	{
		friend class Event<FunctionSignature>;

		Event<FunctionSignature>* m_event;
		boost::function<FunctionSignature> m_func;

		typename Event<FunctionSignature>::ConnectionList::iterator m_connInEventIter;

//		void Invalidate() {m_event = 0;}
		bool IsValid() const
		{
			return m_event != 0;
		}

	public:
		Connection(Event<FunctionSignature>& ev, const boost::function<FunctionSignature>& func);
		Connection(typename Event<FunctionSignature>::Port& evPort, const boost::function<FunctionSignature>& func);
		~Connection();
	};

	// Ports allow Event hosts to pass their clients a handle which the clients can create valid 
	// EventConnections with.  An EventConnection made with an Event's Port is equal to an 
	// EventConnection made with an Event.  Ports allow host classes to shield the Event::Fire()
	// method from their clients (so the clients can't call it, all they can do is connect).
	template <class FunctionSignature>
	class Port
	{
		friend class Event<FunctionSignature>;
		friend class Connection<FunctionSignature>;

		Event<FunctionSignature>& m_hostEvent;
		Port(Event<FunctionSignature>& hostEvent) : m_hostEvent(hostEvent) {};
	};
}

template <class T>
class Event
{
};


// 0 parameters
template <>
class Event<void ()>
{
	typedef void FunctionSignature();
	typedef boost::function<FunctionSignature> FunctorType;

	typedef std::list<EventDetails::Connection<FunctionSignature>*> ConnectionList;
	ConnectionList m_connectionList;

	friend class EventDetails::Connection<FunctionSignature>;

	ConnectionList::iterator Attach(EventDetails::Connection<FunctionSignature>* conn)
	{
		m_connectionList.push_front(conn);
		return m_connectionList.begin();
	}

	void Detach(ConnectionList::iterator& connIter)
	{
		m_connectionList.erase(connIter);
	}

public:
	~Event()
	{
		ASSERTMSG(m_connectionList.empty(), "Don't delete an event before you delete the connections to it");
		/*
		for(ConnectionList::iterator iter = m_connectionList.begin(); iter != m_connectionList.end(); ++iter)
		{
			(*iter)->Invalidate();
		}
		*/
	}

	typedef EventDetails::Connection<FunctionSignature> Connection;
	typedef EventDetails::Port<FunctionSignature> Port;

	Port GetPort() {return Port(*this);}


	void Fire()
	{
		for(ConnectionList::iterator iter = m_connectionList.begin(); iter != m_connectionList.end(); ++iter)
		{
			(*iter)->m_func(); // Call the handlers
		}
	}
};

// 1 parameters
template <class P1>
class Event<void (P1)>
{
	typedef void FunctionSignature(P1);
	typedef boost::function<FunctionSignature> FunctorType;

	typedef std::list<EventDetails::Connection<FunctionSignature>*> ConnectionList;
	ConnectionList m_connectionList;

	friend class EventDetails::Connection<FunctionSignature>;

	typename ConnectionList::iterator Attach(EventDetails::Connection<FunctionSignature>* conn)
	{
		m_connectionList.push_front(conn);
		return m_connectionList.begin();
	}

	void Detach(typename ConnectionList::iterator& connIter)
	{
		m_connectionList.erase(connIter);
	}

public:
	~Event()
	{
		ASSERTMSG(m_connectionList.empty(), "Don't delete an event before you delete the connections to it");
	}

	typedef EventDetails::Connection<FunctionSignature> Connection;
	typedef EventDetails::Port<FunctionSignature> Port;

	Port GetPort() {return Port(*this);}


	void Fire(P1 p1)
	{
		for(typename ConnectionList::iterator iter = m_connectionList.begin(); iter != m_connectionList.end(); ++iter)
		{
			(*iter)->m_func(p1); // Call the handlers
		}
	}
};

// 1 parameters
template <class P1, class P2>
class Event<void (P1, P2)>
{
	typedef void FunctionSignature(P1, P2);
	typedef boost::function<FunctionSignature> FunctorType;

	typedef std::list<EventDetails::Connection<FunctionSignature>*> ConnectionList;
	ConnectionList m_connectionList;

	friend class EventDetails::Connection<FunctionSignature>;

	typename ConnectionList::iterator Attach(EventDetails::Connection<FunctionSignature>* conn)
	{
		m_connectionList.push_front(conn);
		return m_connectionList.begin();
	}

	void Detach(typename ConnectionList::iterator& connIter)
	{
		m_connectionList.erase(connIter);
	}

public:
	~Event()
	{
		ASSERTMSG(m_connectionList.empty(), "Don't delete an event before you delete the connections to it");
	}

	typedef EventDetails::Connection<FunctionSignature> Connection;
	typedef EventDetails::Port<FunctionSignature> Port;

	Port GetPort() {return Port(*this);}


	void Fire(P1 p1, P2 p2)
	{
		for(typename ConnectionList::iterator iter = m_connectionList.begin(); iter != m_connectionList.end(); ++iter)
		{
			(*iter)->m_func(p1, p2); // Call the handlers
		}
	}
};

namespace EventDetails
{
	template <class FunctionSignature>
	Connection<FunctionSignature>::Connection(Event<FunctionSignature>& ev, const boost::function<FunctionSignature>& func)
		: m_event(&ev)
		, m_func(func)
	{
		m_connInEventIter = m_event->Attach(this);
	}
	template <class FunctionSignature>
	Connection<FunctionSignature>::Connection(typename Event<FunctionSignature>::Port& evPort, const boost::function<FunctionSignature>& func)
		: m_event(&evPort.m_hostEvent)
		, m_func(func)
	{
		m_connInEventIter = m_event->Attach(this);
	}

	template <class FunctionSignature>
	Connection<FunctionSignature>::~Connection()
	{
		m_event->Detach(m_connInEventIter);
	}
}

#endif
