#ifndef GRAPHICSFRAME_H_
#define GRAPHICSFRAME_H_

#include <TriangleMesh.h>
#include <Materials.h>
#include <CommandQueue.h>
#include <Algebra.h>
#include <FixedStack.h>
#include <Memory>

struct LineSegment
{
	Vector<4,VT_INSTANCE,false> a;
	Vector<4,VT_INSTANCE,false> b;
};

class GraphicsFrame
{
public:
	class Packet
	{
	public:
		enum RenderObject
		{
			RO_TRIMESH,
			RO_LINE
		};
		RenderObject m_objType;
	private:

		int m_modelView;
		const LineSegment* m_line;
		const TriangleMesh* m_mesh;
		const Material* m_material;

		float m_alpha;

		int m_layer;

		bool m_wireframe;

		float m_lineThickness;

		bool m_lights;

		friend class GraphicsFrame;
		GraphicsFrame* m_frame;
		void AttachToFrame(GraphicsFrame& frame) {m_frame = &frame;}

	public:
		// The modelView integer is an index that is returned by GraphicsFrame::RegisterMatrix
		Packet(int modelView, const TriangleMesh* mesh, const Material* material, float alpha = 1.0f, int layer=0, bool wireframe=false)
			: m_objType(RO_TRIMESH)
			, m_modelView(modelView)
			, m_line(0)
			, m_mesh(mesh)
			, m_material(material)
			, m_alpha((alpha > 1.0f ? 1.0f : alpha)*(material?material->alpha:1.0f))
			, m_frame(0)
			, m_layer(layer)
			, m_wireframe(wireframe)
			, m_lights(true)
		{
		};
		// Constructor to create a line segment packet
		Packet(int modelView, const LineSegment* line, const Material* material, float thickness, float alpha = 1.0f, int layer=0)
			: m_objType(RO_LINE)
			, m_modelView(modelView)
			, m_line(line)
			, m_mesh(0)
			, m_material(material)
			, m_alpha((alpha > 1.0f ? 1.0f : alpha)*(material?material->alpha:1.0f))
			, m_frame(0)
			, m_layer(layer)
			, m_wireframe(false)
			, m_lineThickness(thickness)
			, m_lights(false)
		{};

		const Matrix<4,4,false>& GetModelView() const {ASSERT(m_frame != 0); return m_frame->GetRegisteredMatrix(m_modelView);}
		const TriangleMesh* GetMesh() const {return m_mesh;}
		const Material* GetMaterial() const {return m_material;}

		int GetLayer() const {return m_layer;}

		bool HasAlpha() const {return m_alpha < 1.0f;}
		float GetAlpha() const {ASSERT(HasAlpha()); return m_alpha;}

		bool GetWireframe() const {return m_wireframe;}

		RenderObject GetRenderObjectType() const {return m_objType;}

		const LineSegment* GetLine() const {return m_line;}

		float GetLineThickness() const {return m_lineThickness;}

		bool IsLit() const {return m_lights;}
	};


	struct StateChangeOrder
	{
		bool operator()(const Packet* lsInd, const Packet* rsInd) const;
	};
	typedef std::set< const Packet*, StateChangeOrder > SortedPacketSet;

private:
	FixedStack< Matrix<4,4, false> > m_matrices;
	FixedStack<Packet> m_packets;


	CommandQueue* m_commandQueue;
	SortedPacketSet m_drawSortedPackets;

	Mutex m_sortedPacketsLock;
	Mutex m_freezeLock; // Prevents this class from being destructed before we're ready

	struct AddPacketCommand : public CommandQueue::Command
	{
		GraphicsFrame& handler;
		const Packet* packet;

		AddPacketCommand(GraphicsFrame& pHandler, const Packet* pPacket)
			: handler(pHandler)
			, packet(pPacket)
		{};

		virtual void ProcessCommand() {handler.ProcessCommand(*this);}
	};
	void ProcessCommand(AddPacketCommand& cmd);

	bool m_locked;

	Vector<4> m_lightDir;

	Vector<3> m_backgroundColour;

public:
	GraphicsFrame(unsigned int maxMatrices, unsigned int maxPackets);
	~GraphicsFrame();

	int RegisterMatrix(const Matrix<4,4>& mat)
	{
		m_matrices.Push(mat);
		return m_matrices.NumItems()-1;
	}
	const Matrix<4,4,false>& GetRegisteredMatrix(int matNum) const {return m_matrices[matNum];}

	void AddPacket(const Packet& packet);

	void Lock(); // Locks down the frame, and opens access to the packets to public

	typedef std::auto_ptr<Mutex::ScopedLock> Freeze;
	Freeze MakeFreeze() {return Freeze(new Mutex::ScopedLock(m_freezeLock));}

	// This structure locks the sorted packet lock on construction and releases it on destruction
	const SortedPacketSet& GetPackets() {ASSERT(m_locked); return m_drawSortedPackets;}

	void SetLightDir(const Vector<4>& dir) {m_lightDir = dir;} // Kinda temporary
	const Vector<4>& GetLightDir() const {return m_lightDir;}

	void SetBackgroundColour(const Vector<3>& colour) {m_backgroundColour = colour;}
	const Vector<3>& GetBackgroundColour() const {return m_backgroundColour;}
};

#endif
