#include "FileFetcher.h"
#include "FileFetcherLoadWorker.h"

FileFetcher::FileLoad::FileLoad(FileFetcher& fetcher, const ResourceIdentifier& resId, const boost::function<FileLoadDoneSig>& doneCB)
	: m_fileFetcher(fetcher)
	, m_resId(resId)
	, m_data(0)
	, m_dataSize(-1)
	, m_error(false)
	, m_currentWorkUnit(0)
	, m_positionInLine(fetcher.m_queuedFileLoads.end())
	, m_doneCB(doneCB)
{
	m_fileFetcher.RegisterFileLoad(this);
}
FileFetcher::FileLoad::~FileLoad()
{
	m_fileFetcher.UnregisterFileLoad(this);
}

bool FileFetcher::FileLoad::Done() const
{
	Mutex::ScopedLock lock(m_monitorMutex); 
	return PrivateIsDone();
}
bool FileFetcher::FileLoad::Error() const 
{
	Mutex::ScopedLock lock(m_monitorMutex); 
	ASSERT(PrivateIsDone());
	return m_error;
}

const void* FileFetcher::FileLoad::GetData() const 
{
	Mutex::ScopedLock lock(m_monitorMutex); 
	ASSERT(PrivateIsDone());
	ASSERT(m_data != 0); 
	return m_data;
}

long FileFetcher::FileLoad::GetDataSize() const
{
	Mutex::ScopedLock lock(m_monitorMutex); 
	ASSERT(PrivateIsDone());
	ASSERT(m_data != 0); 
	return m_dataSize;
}

void FileFetcher::FileLoad::StopLoadProcess()
{
	m_fileFetcher.StopLoadProcess(this);
}

FileFetcher::FileFetcher(long int maxMem)
: m_fetcherAllocator(maxMem)
, m_workersToShutdown(0)
{
	StartupWorkers();
}

FileFetcher::~FileFetcher()
{
	ShutdownWorkers();
}

void FileFetcher::RegisterFileLoad(FileLoad* loadToRegister)
{
	Mutex::ScopedLock lock(m_monitorMutex);
	Mutex::ScopedLock fileLock(loadToRegister->m_monitorMutex);

	// We have nobody available to do the load for us, so just queue it up and wait
	loadToRegister->m_positionInLine = m_queuedFileLoads.insert(m_queuedFileLoads.end(), loadToRegister);
	m_fileLoadQueuedCond.SignalAll();
}

void FileFetcher::UnregisterFileLoad(FileLoad* loadToUnregister)
{
	Mutex::ScopedLock lock(m_monitorMutex);
	Mutex::ScopedLock fileLock(loadToUnregister->m_monitorMutex);

	// Also deallocate any memory we were lending out
	if(loadToUnregister->m_data != 0)
	{
		m_fetcherAllocator.Free(loadToUnregister->m_data);
		loadToUnregister->m_data = 0;
	}

	PrivateStopLoadProcess(loadToUnregister);
}

void FileFetcher::StopLoadProcess(FileLoad* loadToStop)
{
	Mutex::ScopedLock lock(m_monitorMutex);
	Mutex::ScopedLock fileLock(loadToStop->m_monitorMutex);

	PrivateStopLoadProcess(loadToStop);
}

void FileFetcher::PrivateStopLoadProcess(FileLoad* loadToStop)
{
	if(loadToStop->m_currentWorkUnit != 0)
	{
		loadToStop->m_currentWorkUnit->fileLoad = 0;
		loadToStop->m_currentWorkUnit = 0;
	}

	if(loadToStop->m_positionInLine != m_queuedFileLoads.end())
	{
		// Remove it from the queue
		m_queuedFileLoads.erase(loadToStop->m_positionInLine);
		loadToStop->m_positionInLine = m_queuedFileLoads.end();
	}
}



FileFetcher::WorkUnit* FileFetcher::RequestWork()
{
	Mutex::ScopedLock lock(m_monitorMutex);

	while(m_queuedFileLoads.empty())
	{
		m_fileLoadQueuedCond.Wait(m_monitorMutex);
	}

	// We have work all ready for you!!  Please accept.
	FileLoad* nextWork = m_queuedFileLoads.front();
	m_queuedFileLoads.pop_front();

	if(nextWork == 0)
	{
		ASSERT(m_workersToShutdown > 0);
		--m_workersToShutdown;

		if(m_workersToShutdown == 0)
		{
			m_allWorkersShutdownCond.SignalAll();
		}

		return 0;
	}


	Mutex::ScopedLock fileLock(nextWork->m_monitorMutex);

	WorkUnit* workUnit = new WorkUnit(nextWork);
	nextWork->m_currentWorkUnit = workUnit;
	nextWork->m_positionInLine = m_queuedFileLoads.end();

	return workUnit;
}


void FileFetcher::ReturnWorkResults(WorkUnit* workResults)
{
	Mutex::ScopedLock lock(m_monitorMutex);

	if(workResults->fileLoad != 0)
	{
		{
			Mutex::ScopedLock fileLock(workResults->fileLoad->m_monitorMutex);

			ASSERT(workResults->fileLoad->m_data == 0);

			// Okay, it's still a valid file load (ie: the original FileLoad object has not been
			// destroyed), so return the results
			if(workResults->data == 0)
			{
				workResults->fileLoad->m_error = true;
			}
			else
			{
				workResults->fileLoad->m_data = workResults->data;
				workResults->fileLoad->m_dataSize = workResults->dataSize;
			}

			ASSERT(workResults->fileLoad->m_currentWorkUnit != 0);
			workResults->fileLoad->m_currentWorkUnit = 0;
		}

		if(!workResults->fileLoad->m_doneCB.empty())
		{
			workResults->fileLoad->m_doneCB(workResults->fileLoad);
		}
	}
	else
	{
		// The nerve!  We go and get this data, and now they don't even want it anymore!
		if(workResults->data != 0)
		{
			m_fetcherAllocator.Free(workResults->data);
		}
	}

	
	delete workResults;
}

void FileFetcher::StartupWorkers()
{
	// Initialize our worker threads
	const int NUM_LOADERS = 4; // This is the maximum number of files we can have queued to load within the OS at a time
	for(int i=0; i<NUM_LOADERS; ++i)
	{
		m_workers.push_back(new FileFetcherLoadWorker(*this));
	}
}

void FileFetcher::ShutdownWorkers()
{
	Mutex::ScopedLock lock(m_monitorMutex);

	m_workersToShutdown = m_workers.size();

	for(int i=0; i<m_workersToShutdown; ++i)
	{
		m_queuedFileLoads.push_back(0);
	}
	m_fileLoadQueuedCond.SignalAll();

	if(m_workersToShutdown > 0)
	{
		m_allWorkersShutdownCond.Wait(m_monitorMutex);
	}

	// They should all be shutdown now, delete them
	while(!m_workers.empty())
	{
		delete m_workers.back();
		m_workers.pop_back();
	}
}
