#include "OBJFile.h"

#include <Exceptions.h>
#include <ResourceManager.h>

#include <ScopedFStream.h>

#include <sstream>
#include <string>

// MTL files are auxillary files referenced by OBJ files
// which contain material information
class MTLFile
{
private:
	SafePointerList<OBJFile::Material> m_materials;

public:
	MTLFile(std::istream& in, ResourceManager& resMan)
	{
		int lineNum = 0;
		OBJFile::Material* curMaterial = 0;

		while(!in.eof())
		{
			++lineNum;

			char lineBuffer[500];
			in.getline(lineBuffer, 500);

			std::istringstream liness(lineBuffer);

			std::string command;
			liness >> command;

			if(command.size() == 0) continue; // We don't deal blank lines
			if(command[0] == '#') continue; // A comment (not me, him)

			if(command == "newmtl")
			{
				char restBuf[500];
				liness.getline(restBuf, 500);

				m_materials.objs.push_back(new OBJFile::Material());
				curMaterial = m_materials.objs.back();
				
				curMaterial->name = std::string(restBuf).substr(1);
			}
			else if(command == "Ns")
			{
				if(curMaterial == 0) throw ParseException(lineNum, "No material has been defined yet!");

				liness >> curMaterial->shininess;
			}
			else if(command == "Ka")
			{
				if(curMaterial == 0) throw ParseException(lineNum, "No material has been defined yet!");

				liness >> curMaterial->ambientColour[0];
				liness >> curMaterial->ambientColour[1];
				liness >> curMaterial->ambientColour[2];
			}
			else if(command == "Kd")
			{
				if(curMaterial == 0) throw ParseException(lineNum, "No material has been defined yet!");

				liness >> curMaterial->diffuseColour[0];
				liness >> curMaterial->diffuseColour[1];
				liness >> curMaterial->diffuseColour[2];
			}
			else if(command == "Ks")
			{
				if(curMaterial == 0) throw ParseException(lineNum, "No material has been defined yet!");

				liness >> curMaterial->specularColour[0];
				liness >> curMaterial->specularColour[1];
				liness >> curMaterial->specularColour[2];
			}
			else if(command == "d")
			{
				if(curMaterial == 0) throw ParseException(lineNum, "No material has been defined yet!");

				liness >> curMaterial->alpha;
			}
			else if(command == "illum")
			{
				if(curMaterial == 0) throw ParseException(lineNum, "No material has been defined yet!");

				int illumType;
				liness >> illumType;

				if(illumType == 1)
				{
					curMaterial->useSpecular = false;
				}
				else if(illumType == 2)
				{
					curMaterial->useSpecular = true;
				}
				else
				{
					throw ParseException(lineNum, "Invalid illum value (currently only support 1 and 2)");
				}
			}


		}
	}

	int GetNumMaterials() const {return m_materials.objs.size();}
	const OBJFile::Material& GetMaterial(int index) const {return *m_materials.objs[index];}
};

OBJFile::OBJFile(std::istream& in, ResourceManager& resMan)
{
	int lineNum = 0;

	Object* curObject = 0;
	Material* curMaterial = 0;

	// Parse the input file and save the data
	while(!in.eof())
	{
		++lineNum;

		char lineBuffer[500];
		in.getline(lineBuffer, 500);

		std::istringstream liness(lineBuffer);

		std::string command;
		liness >> command;

		if(command.size() == 0) continue; // We don't deal blank lines
		if(command[0] == '#') continue; // A comment (not me, him)

		if(command == "o")
		{
			m_objectList.objs.push_back(new Object());
			curObject = m_objectList.objs.back();

			bool first = true;
			while(!liness.eof())
			{
				if(first) first = false;
				else curObject->name += " ";

				std::string chunk;
				liness >> chunk;
				curObject->name += chunk;
			}
		}
		else if(command == "v")
		{
			if(curObject == 0) throw ParseException(lineNum, "Read vertex, but no object declared yet");

			float elems[3];
			liness >> elems[0];
			liness >> elems[1];
			liness >> elems[2];

			curObject->vertices.push_back(Vertex(elems));
		}
		else if(command == "vn")
		{
			if(curObject == 0) throw ParseException(lineNum, "Reading normal, but no object declared yet");
				
			float elems[3];
			liness >> elems[0];
			liness >> elems[1];
			liness >> elems[2];

			curObject->normals.push_back(Vertex(elems));
		}
		else if(command == "f")
		{
			if(curObject == 0) throw ParseException(lineNum, "Read face, but no object declared yet");

			Face face;
			face.numVertices = 0;

			while(!liness.eof())
			{
				std::string vertInfo;
				liness >> vertInfo;
				if(vertInfo.size() == 0) break;

				std::istringstream vertss(vertInfo);
				for(int i=0; i<3 && !vertss.eof(); ++i)
				{
					char buf[500];
					vertss.getline(buf, 500, '/');
					std::istringstream numss(buf);

					if(std::string(buf).size() == 0) continue;

					switch(i)
					{
						case 0: // vertices
						{
							numss >> face.vertexIndices[face.numVertices];
							--face.vertexIndices[face.numVertices]; // OBJ file has index starting at 1, not 0
						} break;

						case 1: // uv coordinates
						{
						} break;

						case 2: // normals
						{
							numss >> face.normalIndices[face.numVertices];
							--face.normalIndices[face.numVertices];
						} break;
					};
				};

				++face.numVertices;
			};

			if(face.numVertices != 3 && face.numVertices != 4)
			{
				throw ParseException(lineNum, "Reading a face that has an unsupported number of vertices");
			}

			face.material = curMaterial;

			curObject->faces.push_back(face);
		}
		else if(command == "mtllib")
		{
			char restLine[500];
			liness.getline(restLine, 500);
			std::string matFilename = std::string(restLine).substr(1);

			// Load in a set of new materials
			ResourceReference<MTLFile> matLib(resMan, ResourceIdentifier(matFilename));

			// Add all the defined materials in to our own list
			for(int i=0; i<matLib->GetNumMaterials(); ++i)
			{
				m_materialList.objs.push_back(new Material(matLib->GetMaterial(i)));
				if(m_materials.find(matLib->GetMaterial(i).name) != m_materials.end())
				{
					std::ostringstream oss;
					oss << "The material '" << matLib->GetMaterial(i).name << "' is being defined twice";
					throw ParseException(lineNum, oss.str());
				}

				m_materials[matLib->GetMaterial(i).name] = m_materialList.objs.back();
			}
		}
		else if(command == "usemtl")
		{
			char restLine[500];
			liness.getline(restLine, 500);
			std::string matToUse = std::string(restLine).substr(1);

			MaterialMap::const_iterator found = m_materials.find(matToUse);
			if(found == m_materials.end())
			{
				curMaterial = 0;
				/*
				std::ostringstream oss;
				oss << "The material '" << matToUse << "' has not been defined yet";
				throw ParseException(lineNum, oss.str());
				*/
			}
			else
			{
				curMaterial = found->second;
			}
		}
	};

	// Okay, we made it, now set transfer ownership of our objects to the class
	for(std::vector<Object*>::iterator iter = m_objectList.objs.begin(); iter != m_objectList.objs.end(); ++iter)
	{
		m_objects[(*iter)->name] = *iter;
	}
}


OBJFile::~OBJFile()
{
}
