#include <OSWindow.h>
#include <GraphicsContext.h>
#include <MouseInput.h>
#include <KeyboardInput.h>

#include <System.h>
#include <R3Utils.h>

#include <Model.h>
#include <OBJFile.h>
#include <Materials.h>

#include <ArcBall.h>
#include <Projection.h>

#include <ScopedFStream.h>
#include <Exceptions.h>

#include <math.h>
#include <iostream>
#include <boost/bind.hpp>
#include <fstream>

#include <float.h>

static bool s_toggleFullscreen = false;

static bool s_quitFlag = false;

static bool s_dragging = false;

static Mutex s_rotLock;

static UnitQuaternion s_ballRot;

static GenericWindow::MousePosition s_currMousePos(0,0);
static GenericWindow::MousePosition s_lastMousePos(0,0);

static void ProcessMouseMove(OSWindow* win, const MouseInput::MoveEvent& moveEvent)
{
	if(s_dragging)
	{
		Mutex::ScopedLock lock(s_rotLock);

		s_lastMousePos = s_currMousePos;
		s_currMousePos = win->GetLastMousePosition();

		float minDim = (win->GetWidth() < win->GetHeight() ? win->GetWidth() : win->GetHeight());
		minDim *= 0.7;

		float centeredCurrX = s_currMousePos.x - win->GetWidth()/2;
		float centeredCurrY = s_currMousePos.y - win->GetHeight()/2;
		float centeredLastX = s_lastMousePos.x - win->GetWidth()/2;
		float centeredLastY = s_lastMousePos.y - win->GetHeight()/2;

		Vector<2> from(2.0f * centeredLastX/minDim, -(2.0f * centeredLastY/minDim));
		Vector<2> to(2.0f * centeredCurrX/minDim, -(2.0f * centeredCurrY/minDim));

		UnitQuaternion rotQuat = GetRotationFromArcBallPointDelta(from, to);
		s_ballRot = rotQuat*s_ballRot;
	}
}

static void ProcessMouseButton(OSWindow* win, const MouseInput::ButtonEvent& buttonEvent)
{
	if(buttonEvent.button == 0)
	{
		Mutex::ScopedLock lock(s_rotLock);
		if(buttonEvent.pressed)
		{
			s_currMousePos = win->GetLastMousePosition();
			s_lastMousePos = s_currMousePos;
		}

		s_dragging = buttonEvent.pressed;
	}
}

static void MarkQuitFlag()
{
	s_quitFlag = true;
}

int main(int argc, char** args)
{
	OSWindow mainWin("ArcBall Test", 1024, 768);

	mainWin.AttachQuitEventHandler(&MarkQuitFlag);

	MouseInput::MoveListenerConnection mouseListener(mainWin.GetMouseInput(), boost::bind(&ProcessMouseMove, &mainWin, _1));
	MouseInput::ButtonListenerConnection mouseButtonListener(mainWin.GetMouseInput(), boost::bind(&ProcessMouseButton, &mainWin, _1));

	GenericWindow::QuitEventConnection quitEventConn = mainWin.AttachQuitEventHandler(&MarkQuitFlag);

	Vector<4> lightPosVec(0, 0, 1.0f, 0);

	try
	{
		MaterialSet materials;
		OBJFile testObj(ScopedFStream<std::ifstream>("UnitSphere.obj").s, ".");
		Model testBox(testObj.GetObjectItem(0), AddMaterialsFromOBJFile(materials, testObj));

		float curSec = System::GetTimeInSeconds();
		while(!s_quitFlag)
		{
			if(s_toggleFullscreen)
			{
				s_toggleFullscreen = false;
				mainWin.SetWindowMode(mainWin.GetWindowMode() == OSWindow::M_WINDOWED ? OSWindow::M_FULLSCREEN : OSWindow::M_WINDOWED);
			}

			mainWin.GetGraphicsContext().FlushWait();

			GraphicsFrame frame(1000, 1000);
			frame.SetLightDir(lightPosVec);

			float nextSec = System::GetTimeInSeconds();
			float deltaT = nextSec - curSec;
			curSec = nextSec;

			Matrix<4,4> modelView(Matrix<4,4>::Identity);
			modelView.SetColumn(3, Vector<4>(0,0,-5,1));

			{
				Mutex::ScopedLock lock(s_rotLock);

				modelView = modelView * s_ballRot.ToMatrix();
			}

			modelView = modelView * MakeUniScaleMatrix(0.7f);
			testBox.AddToGraphicsFrame(frame, modelView);

			frame.Lock();

			float renderWidth=1.0f;
			float renderHeight=1.0f;
			if(mainWin.GetWidth() < mainWin.GetHeight())
			{
				renderWidth = 2.0f;
				renderHeight = 2.0f * mainWin.GetHeight() / mainWin.GetWidth();
			}
			else
			{
				renderWidth = 2.0f * mainWin.GetWidth() / mainWin.GetHeight();
				renderHeight = 2.0f;
			}

			mainWin.GetGraphicsContext().RenderGraphicsFrame(frame, GraphicsContext::RenderOptions(MakeOrthographicProjectionMatrix(-renderWidth/2, renderWidth/2, -renderHeight/2, renderHeight/2, 0, 100.0f)));
		}
	}
	catch(const ParseException& e)
	{
		std::cerr << "Parse error: " << e.GetMsg() << std::endl;
	}
	catch(ScopedFStream<std::ifstream>::IOException& e)
	{
		std::cerr << "IO error: " << e.GetMsg() << std::endl;
	}

	return 0;
}

