#include <FileFetcher.h>
#include <ResourceCreator.h>

#include <BoundedQueue.h>

#include <iostream>

class Foo
{
	char str[10];

public:
	Foo(std::istream& in)
	{
		for(int i=0; i<10; ++i) str[i] = in.get();
	}

	const char* GetStr() const {return str;}
};

class Roo
{
	int num;

public:
	Roo(std::istream& in)
	{
		in >> num;
	}

	int GetNum() const {return num;}
};


// A goo file format is a number followed by that many lines, where
// each line is a filename to another goo file (not the same one though,
// or else we'd have an endless loop).  Eventually, one goo will have to feature
// a first line of "0", so as to be the base case.
class Goo
{
	static const int MAX_GOOS = 10;
	int m_numExtras;
	Goo* m_extraGoos[MAX_GOOS];

	Mutex m_constructionMutex;
	Condition m_auxGooLoadedCond;

	TypedResourceFactory<Goo>& m_factory;

	BoundedQueue<int> m_gooLoadedQueue;

	void AuxillaryGooLoaded(int gooNum, ResourceCreator::CreationProcess* creationProcess, void* data)
	{
		ASSERT(data != 0); // Error loading auxillary goo
		m_extraGoos[gooNum] = reinterpret_cast<Goo*>(data);

		std::cout << "Auxillary goo loaded!" << std::endl;

		m_gooLoadedQueue.Push(gooNum);
	}

public:
	Goo(std::istream& in, ResourceCreator& resCreator, TypedResourceFactory<Goo>& factory)
		: m_factory(factory)
		, m_gooLoadedQueue(MAX_GOOS)
	{
		in >> m_numExtras;
		ASSERT(m_numExtras <= MAX_GOOS);

		// Get rid of the newline
		char temp[100];
		in.getline(temp, sizeof(temp)-1);

		std::set<int> waitingList;

		ResourceCreator::CreationProcess* creations[MAX_GOOS];
		for(int i=0; i<m_numExtras; ++i)
		{
			char filenameBuf[256];

			m_extraGoos[i] = 0;

			in.getline(filenameBuf, sizeof(filenameBuf)-1);
			if(filenameBuf[0] != 0 && filenameBuf[strlen(filenameBuf)-1] == 13)
			{
				// Kill the trailing newline
				filenameBuf[strlen(filenameBuf)-1] = 0;
			}

			// Load up our extra goos now
			creations[i] = new ResourceCreator::CreationProcess(resCreator, ResourceIdentifier(filenameBuf), m_factory, boost::bind(&Goo::AuxillaryGooLoaded, this, i, _1, _2));

			waitingList.insert(i);
		}

		std::set<int>::iterator waitingOn;
		if(!waitingList.empty())
		{
			waitingOn = waitingList.begin();
			creations[*waitingOn]->MarkWaitedOn(Thread::GetCurrentID());
		}

		for(int i=0; i<m_numExtras; ++i)
		{
			int gooLoaded = m_gooLoadedQueue.Pop();

			if(gooLoaded == *waitingOn)
			{
				creations[gooLoaded]->MarkNotWaitedOn(Thread::GetCurrentID());
				waitingOn = waitingList.end();
			}
			waitingList.erase(gooLoaded);
	
			if(!waitingList.empty() && waitingOn == waitingList.end())
			{
				waitingOn = waitingList.begin();
				creations[*waitingOn]->MarkWaitedOn(Thread::GetCurrentID());
			}
		}

		for(int i=0; i<m_numExtras; ++i)
		{
			delete creations[i];
		}
	}

	~Goo()
	{
		for(int i=0; i<m_numExtras; ++i)
		{
			delete m_extraGoos[i];
		}
	}

	int GetNumExtras() const {return m_numExtras;}
	const Goo* GetExtraGoo(int index) const {ASSERT(index >= 0 && index < m_numExtras); return m_extraGoos[index];}
};

template <class T>
class GenericAllocator : public TypedResourceFactory<T>
{
public:
	virtual void CreateObject(void* mem, std::istream& in)
	{
		new(mem) T(in);
	}
};

// Foo gets a special factory because it takes constructor parameters
class GooFactory : public TypedResourceFactory<Goo>
{
	ResourceCreator& m_resCreator;

public:
	GooFactory(ResourceCreator& resCreator)
		: m_resCreator(resCreator)
	{};
	virtual void CreateObject(void* mem, std::istream& in)
	{
		new(mem) Goo(in, m_resCreator, *this);
	}
};

static GenericAllocator<Foo> s_fooFactory;
static GenericAllocator<Roo> s_rooFactory;

static void FooDoneCallback(ResourceCreator::CreationProcess* creationProcess, void* data)
{
	if(data == 0)
	{
		std::cout << "Foo fail!" << std::endl;
		return;
	}

	Foo* foo = (Foo*)data;

	// Check that foo loaded correctly
	const char* testString = "Foo!";
	ASSERT(memcmp(foo->GetStr(), testString, strlen(testString)) == 0);

	std::cout << "Foo Pass!" << std::endl;

	delete foo;
}
static void RooDoneCallback(ResourceCreator::CreationProcess* creationProcess, void* data)
{
	if(data == 0)
	{
		std::cout << "Roo fail!" << std::endl;
		return;
	}

	Roo* roo = (Roo*)data;

	ASSERT(roo->GetNum() == 45678);

	std::cout << "Roo Pass!" << std::endl;

	delete roo;
}
static void GooDoneCallback(ResourceCreator::CreationProcess* creationProcess, void* data)
{
	if(data == 0)
	{
		std::cout << "Goo fail!" << std::endl;
		return;
	}

	Goo* goo = (Goo*)data;

	ASSERT(goo->GetNumExtras() == 3);
	for(int i=0; i<2; ++i)
	{
		const Goo* nextGoo = goo->GetExtraGoo(i);
		ASSERT(nextGoo->GetNumExtras() == 2);
		for(int j=0; j<2; ++j)
		{
			ASSERT(nextGoo->GetExtraGoo(j)->GetNumExtras() == 0);
		}
	}
	ASSERT(goo->GetExtraGoo(2)->GetNumExtras() == 0);

	std::cout << "Full goo loaded!" << std::endl;
}

void TestThread(ResourceCreator* resCreator, int threadId)
{
	srand(threadId);

	const int NUM_ITERATIONS = 50;
	const int MAX_SLEEP_TIME = 1; // In seconds

	GooFactory gooFactory(*resCreator);

	for(int i=0; i<NUM_ITERATIONS; ++i)
	{
		int loadType = rand() % 3;

		const char* fileNames[] =
		{
			"fooTest.txt",
			"rooTest.txt",
			"gooTest.txt"
		};
		ResourceFactory* factories[] =
		{
			&s_fooFactory,
			&s_rooFactory,
			&gooFactory,
		};
		typedef void (*DoneCallback)(ResourceCreator::CreationProcess*, void*);
		DoneCallback callbacks[] =
		{
			&FooDoneCallback,
			&RooDoneCallback,
			&GooDoneCallback,
		};

		ResourceCreator::CreationProcess creation(*resCreator, ResourceIdentifier(fileNames[loadType]), *factories[loadType], callbacks[loadType]);

		Sleep(rand()%(MAX_SLEEP_TIME*1000));
	}
}

int main(int argc, char** args)
{
	StandardAllocator standardAlloc;
	LockingAllocatorDecorator lockingAlloc(standardAlloc);

	FileFetcher fetcher(20000000);
	ResourceCreator resCreator(fetcher, lockingAlloc);

	const int NUM_LOADER_THREADS = 100;
	Thread* threads[NUM_LOADER_THREADS];
	for(int i=0; i<NUM_LOADER_THREADS; ++i)
	{
		threads[i] = new Thread(boost::bind(&TestThread, &resCreator, i));
	}

	for(int i=0; i<NUM_LOADER_THREADS; ++i)
	{
		delete threads[i];
	}

	std::cout << "Done." << std::endl;

	return 0;
}
