#include <ResourceManager.h>

#include <iostream>

class Foo
{
	char str[10];

public:
	Foo(std::istream& in, ResourceManager& resMan)
	{
		std::cout << "Constructing foo" << std::endl;
		for(int i=0; i<10; ++i) str[i] = in.get();
	}
	~Foo()
	{
		std::cout << "Destructing foo" << std::endl;
	}
	const char* GetStr() const {return str;}
};

class Roo
{
	int num;

public:
	Roo(std::istream& in, ResourceManager& resMan)
	{
		std::cout << "Constructing roo" << std::endl;
		in >> num;
	}

	~Roo()
	{
		std::cout << "Destructing roo" << std::endl;
	}

	int GetNum() const {return num;}
};


// A goo file format is a number followed by that many lines, where
// each line is a filename to another goo file (not the same one though,
// or else we'd have an endless loop).  Eventually, one goo will have to feature
// a first line of "0", so as to be the base case.
class Goo
{
	static const int MAX_GOOS = 10;
	int m_numExtras;
	ResourceReference<Goo>* m_extraGoos[MAX_GOOS];

public:
	Goo(std::istream& in, ResourceManager& resMan)
	{
		std::cout << "Constructing goo" << std::endl;

		in >> m_numExtras;
		ASSERT(m_numExtras <= MAX_GOOS);

		// Get rid of the newline
		char temp[100];
		in.getline(temp, sizeof(temp)-1);

		for(int i=0; i<m_numExtras; ++i)
		{
			char filenameBuf[256];

			m_extraGoos[i] = 0;

			in.getline(filenameBuf, sizeof(filenameBuf)-1);
			if(filenameBuf[0] != 0 && filenameBuf[strlen(filenameBuf)-1] == 13)
			{
				// Kill the trailing newline
				filenameBuf[strlen(filenameBuf)-1] = 0;
			}

			// Load up our extra goos now
			m_extraGoos[i] = new ResourceReference<Goo>(resMan, ResourceIdentifier(filenameBuf));
		}

		for(int i=0; i<m_numExtras; ++i)
		{
			m_extraGoos[i]->WaitUntilLoaded();
		}
	}

	~Goo()
	{
		std::cout << "Destructing goo" << std::endl;

		for(int i=0; i<m_numExtras; ++i)
		{
			delete m_extraGoos[i];
		}
	}

	int GetNumExtras() const {return m_numExtras;}
	const Goo& GetExtraGoo(int index) const {ASSERT(index >= 0 && index < m_numExtras); return **m_extraGoos[index];}
};

static void TestFoo(const Foo& foo)
{
	// Check that foo loaded correctly
	const char* testString = "Foo!";
	ASSERT(memcmp(foo.GetStr(), testString, strlen(testString)) == 0);

	std::cout << "Foo Pass!" << std::endl;
}

static void TestRoo(const Roo& roo)
{
	ASSERT(roo.GetNum() == 45678);

	std::cout << "Roo Pass!" << std::endl;
}

static void TestGoo(const Goo& goo)
{
	ASSERT(goo.GetNumExtras() == 3);
	for(int i=0; i<2; ++i)
	{
		const Goo& nextGoo = goo.GetExtraGoo(i);
		ASSERT(nextGoo.GetNumExtras() == 2);
		for(int j=0; j<2; ++j)
		{
			ASSERT(nextGoo.GetExtraGoo(j).GetNumExtras() == 0);
		}
	}
	ASSERT(goo.GetExtraGoo(2).GetNumExtras() == 0);

	std::cout << "Full goo loaded!" << std::endl;
}

void TestThread(ResourceManager* resMan, int threadId)
{
	srand(threadId);

	const int NUM_ITERATIONS = 10;
	const int MAX_SLEEP_TIME = 1; // In seconds

	for(int i=0; i<NUM_ITERATIONS; ++i)
	{
		int loadType = rand() % 3;

		switch(loadType)
		{
		case 0:
			{
				ResourceReference<Foo> fooRes(*resMan, ResourceIdentifier("fooTest.txt"));
				Sleep(rand()%(MAX_SLEEP_TIME*1000));
				TestFoo(*fooRes);
				ASSERT(fooRes.IsLoaded());
			} break;

		case 1:
			{
				ResourceReference<Roo> rooRes(*resMan, ResourceIdentifier("rooTest.txt"));
				Sleep(rand()%(MAX_SLEEP_TIME*1000));
				TestRoo(*rooRes);
				ASSERT(rooRes.IsLoaded());
			} break;

		case 2:
			{
				ResourceReference<Goo> gooRes(*resMan, ResourceIdentifier("gooTest.txt"));
				Sleep(rand()%(MAX_SLEEP_TIME*1000));
				TestGoo(*gooRes);
				ASSERT(gooRes.IsLoaded());
			} break;

		default: ASSERT(false); break;
		};
	}
}

void DoTest(FileFetcher& fetcher, int numThreads, int memSize)
{
	ResourceManager resMan(g_standardAllocator, memSize, fetcher);

	Thread** threads = new Thread*[numThreads];
	for(int i=0; i<numThreads; ++i)
	{
		threads[i] = new Thread(boost::bind(&TestThread, &resMan, i));
	}

	for(int i=0; i<numThreads; ++i)
	{
		delete threads[i];
	}

	delete threads;
}

void DoDestructorTest()
{
	const float MAX_SLEEP_TIME = 0.05; // In seconds

	const int NUM_ITERATIONS = 200;

	for(int i=0; i<NUM_ITERATIONS; ++i)
	{
		FileFetcher fetcher(20000000);
		ResourceManager resMan(g_standardAllocator, 10000, fetcher);

		ResourceReference<Foo> fooRes1(resMan, ResourceIdentifier("fooTest.txt"));
		ResourceReference<Goo> gooRes1(resMan, ResourceIdentifier("gooTest.txt"));
		ResourceReference<Roo> rooRes1(resMan, ResourceIdentifier("rooTest.txt"));
		ResourceReference<Foo> fooRes2(resMan, ResourceIdentifier("fooTest.txt"));
		ResourceReference<Foo> fooRes3(resMan, ResourceIdentifier("fooTest.txt"));
		
		Sleep(rand()%(int)(MAX_SLEEP_TIME*1000));

		std::cout << "Iterate " << i << "/" << NUM_ITERATIONS << std::endl;
	}
}

int main(int argc, char** args)
{
	DoDestructorTest();

	{
		FileFetcher fetcher(20000000);

		DoTest(fetcher, 3, (sizeof(Foo)+sizeof(Roo)+sizeof(Goo)*3));
	}
	{
		FileFetcher fetcher(20000000);

		DoTest(fetcher, 1, (sizeof(Foo)+sizeof(Roo)+sizeof(Goo)*3));
		DoTest(fetcher, 1, (sizeof(Goo)*3));
		DoTest(fetcher, 2, (sizeof(Foo)+sizeof(Roo)+sizeof(Goo)*3));
		DoTest(fetcher, 2, (sizeof(Foo)+sizeof(Goo)*3));
		DoTest(fetcher, 3, (sizeof(Foo)+sizeof(Roo)+sizeof(Goo)*3));
		DoTest(fetcher, 3, (sizeof(Foo)+sizeof(Roo)+sizeof(Goo)*3));

		DoTest(fetcher, 100, (sizeof(Foo)+sizeof(Roo)+sizeof(Goo)*3));
		DoTest(fetcher, 100, (sizeof(Foo)+sizeof(Roo)+sizeof(Goo)*3));
		DoTest(fetcher, 100, (sizeof(Foo)+sizeof(Roo)+sizeof(Goo)*3));
	}

	std::cout << "Done." << std::endl;
	return 0;
}
