#include <ThreadPrimitives.h>
#include <stdlib.h>

#include <stdio>

#include <boost/bind.hpp>

///////////////////////////////////////////////////////////////////////////////
// Test 1
// This test ensures that basic thread creation, execution, and joining are functioning correctly.
// No synchronization objects are used in this test.
static const int NUM_CALCULATIONS = 50;
static int s_calculations[NUM_CALCULATIONS];
static void DoCalculation(int index)
{
	s_calculations[index] = index*index*index;
}

static void SimpleThreadTest()
{
	// Reset the calculations array
	for(int i=0; i<NUM_CALCULATIONS; ++i)
	{
		s_calculations[i] = -1;
	}

	Thread* threads[NUM_CALCULATIONS]; // One thread per calculation

	// Start executing the threads
	for(int i=0; i<NUM_CALCULATIONS; ++i)
	{
		threads[i] = new Thread(boost::bind<void>(&DoCalculation, i));
	}

	// Lalalala... threads executing.

	// Okay, join them now.
	for(int i=0; i<NUM_CALCULATIONS; ++i)
	{
		delete threads[i];
		
		// Check the results
		ASSERT(s_calculations[i] == i*i*i);
	}
}



///////////////////////////////////////////////////////////////////////////////
// Test 2
// This test ensures that basic mutexes are functioning correctly.  We will have
// multiple threads attempting to manipulate one object.  The operations performed
// on the object will be commutative, so it doesn't matter what order they are performed
// in, and so we can check that the final result will be correct.
static const int NUM_ITERATIONS_PER_THREAD = 1000;
static const int NUM_TEST2_THREADS = 20;
static int s_testData;
static Mutex s_test2Mutex;
static void ManipulateTestData()
{
	// Apply an operation to the data 10 times
	for(int i=0; i<NUM_ITERATIONS_PER_THREAD; ++i)
	{
		// Lock the test data before executing computations with it
		Mutex::ScopedLock lock(s_test2Mutex);

		int origTestData = s_testData;

		s_testData = -s_testData;
		s_testData -= 5;

		s_testData += 2 * (origTestData + 5);

		// These calculations amount to: s_testData(new) = s_testData(old) + 5
	}

}

static void SimpleMutexTest()
{
	s_testData = 0;

	Thread* threads[NUM_TEST2_THREADS];

	for(int i=0; i< NUM_TEST2_THREADS; ++i)
	{
		threads[i] = new Thread(&ManipulateTestData);
	}

	for(int i=0; i< NUM_TEST2_THREADS; ++i)
	{
		delete threads[i];
	}

	// Ensure that the result is correct
	ASSERT(s_testData == 5 * NUM_TEST2_THREADS * NUM_ITERATIONS_PER_THREAD);
}

///////////////////////////////////////////////////////////////////////////////
// Test 3
// This test checks that multiple mutexes work fine together.
// We create some arrays of unsigned integers, and then have multiple threads work 
// together to decrement the arrays down to 0, each time increasing one single other
// variable by 5.

static const int TEST3_NUM_ARRAYS = 30;
static int s_intArray[TEST3_NUM_ARRAYS];
static Mutex s_mutexArrays[TEST3_NUM_ARRAYS]; // One mutex for each integer
static int s_test3Sum;
static Mutex s_test3SumMutex;

static const int TEST3_NUM_THREADS = 5;
static void ProcessArray()
{
	while(true)
	{
		bool didSomething = false;

		for(int i=0; i<TEST3_NUM_ARRAYS; ++i)
		{
			Mutex::ScopedLock arrayLock(s_mutexArrays[i]);

			if(s_intArray[i] > 0)
			{
				didSomething = true;

				--s_intArray[i];

				Mutex::ScopedLock sumLock(s_test3SumMutex);

				// s_test3SumMutex += 5
				for(int j=0; j<5; ++j)
				{
					++s_test3Sum; 
				}
			}
		}

		if(!didSomething) break;
	}
}

static void MoreComplicatedMutexTest()
{
	// Initialize our test sum
	s_test3Sum = 0;

	// Initialize our test arrays (and remember what we set them to)
	int sumVals = 0;
	for(int i=0; i<TEST3_NUM_ARRAYS; ++i)
	{
		s_intArray[i] = rand() % 100;
		sumVals += s_intArray[i];
	}

	// Now start up the threads
	Thread* threads[TEST3_NUM_THREADS];
	for(int i=0; i<TEST3_NUM_THREADS; ++i)
	{
		threads[i] = new Thread(&ProcessArray);
	}

	// Lalalala...

	for(int i=0; i<TEST3_NUM_THREADS; ++i)
	{
		delete threads[i];
	}

	// Check the result
	ASSERT(s_test3Sum == sumVals * 5);

	// Check that the arrays are 0
	for(int i=0; i<TEST3_NUM_ARRAYS; ++i)
	{
		ASSERT(s_intArray[i] == 0);
	}
}

///////////////////////////////////////////////////////////////////////////////
// Test 4
// This test ensures that Condition::SignalAll() functionality is working properly,
// because it was not tested in test 4.
// We test this by making a certain amount of worker threads, and then have them repeatedly
// do work, and then wait at a barrier for all the othre threads to finish their work.  We
// check that the threads did their work properly, and then call SignalAll to let them all
// iterate their work again.
static int s_workersWaiting;
static Condition s_workerWaitingCond;
static Condition s_workSignal;
static Mutex s_guardMutex;
static bool s_shouldQuit;

static int s_timesSet[10];
static void WorkerThread(bool* workValue, int myId)
{
	// We signal s_workerWaiting when we're finished working, and we wait for s_workSignal before
	// starting more work.  "Work" is the act of setting *workValue to true

	s_timesSet[myId] = 0;
	while(true)
	{
		{
			// Signal we're ready for work, and then wait for work to arrive
			Mutex::ScopedLock lock(s_guardMutex);
			++s_workersWaiting;
			s_workerWaitingCond.SignalOne();
			s_workSignal.Wait(s_guardMutex);
		}

		if(s_shouldQuit) break;

		ASSERT(*workValue == false);

		*workValue = true;
		++s_timesSet[myId];
	}
}

static void ConditionVariableSignalAllTest()
{
	const int NUM_THREADS = 10;
	bool workValues[NUM_THREADS];
	Thread* threads[NUM_THREADS];

	s_workersWaiting = 0;
	s_shouldQuit = false;

	// Reset work values
	for(int i=0; i<NUM_THREADS; ++i)
	{
		threads[i] = new Thread(boost::bind<void>(&WorkerThread, &workValues[i], i));
	}

	// Reset the work values so that the threads can do their thing
	for(int i=0; i<NUM_THREADS; ++i)
	{
		workValues[i] = false;
	}

	// Wait for all threads to be ready for work
	{
		Mutex::ScopedLock lock(s_guardMutex);

		while(s_workersWaiting < NUM_THREADS)
		{
			s_workerWaitingCond.Wait(s_guardMutex);
		}
	}

	// Our threads should be created, and waiting for work now.  Start doing the tests
	const int NUM_TEST_ITERATIONS = 200;
	for(int j=0; j<NUM_TEST_ITERATIONS; ++j)
	{
		// All threads are waiting for work now
		ASSERT(s_workersWaiting == NUM_THREADS);

		// Make sure we have all all the values reset here
		for(int i=0; i<NUM_THREADS; ++i)
		{
			ASSERT(workValues[i] == false);
		}

		{
			Mutex::ScopedLock lock(s_guardMutex);

			s_workersWaiting = 0;

			// Go!!
			s_workSignal.SignalAll();

			// Wait for the threads to finish
			while(s_workersWaiting < NUM_THREADS)
			{
				s_workerWaitingCond.Wait(s_guardMutex);
			}
		}

		// Check that the values were set correctly, and reset them for the next iteration
		for(int i=0; i<NUM_THREADS; ++i)
		{
			ASSERT(workValues[i] == true);
			workValues[i] = false;
		}
	}

	// Okay we're done, tell the threads to quit now
	s_shouldQuit = true;
	s_workSignal.SignalAll();

	for(int i=0; i<NUM_THREADS; ++i)
	{
		delete threads[i];
	}
}

int main(int argc, char** args)
{
	// Run the tests 5 times because they do not execute exactly the same each time
	for(int i=0; i<5; ++i)
	{
		SimpleThreadTest();

		SimpleMutexTest();
		MoreComplicatedMutexTest();

		ConditionVariableSignalAllTest();

		std::cout << "Successfully executed all tests in pass " << i << std::endl;
	}

	return 0;
}
