#include "RenderAnalyzer.h"

void MakeTransformedCubePoints(const Matrix<4,4>& trans, Vector<4> cubePoints[8])
{
	cubePoints[0] = trans.GetColumn(3); // (0,0,0)
	cubePoints[1] = (trans.GetColumn(0) + trans.GetColumn(3)); // (1,0,0)
	cubePoints[2] = (cubePoints[1] + trans.GetColumn(1)); // (1,1,0)
	cubePoints[3] = (trans.GetColumn(1) + trans.GetColumn(3)); // (0,1,0)
	cubePoints[4] = (cubePoints[3] + trans.GetColumn(2)); // (0,1,1)
	cubePoints[5] = (trans.GetColumn(2) + trans.GetColumn(3)); // (0,0,1)
	cubePoints[6] = (trans.GetColumn(0) + cubePoints[5]); // (1,0,1)
	cubePoints[7] = (trans.GetColumn(1) + cubePoints[6]); // (1,1,1)
}

static void MakeScreenCubePoints(const Matrix<4,4>& modelScreen, Vector<4> cubePoints[8], float origZs[8], float nearZ)
{
	MakeTransformedCubePoints(modelScreen, cubePoints);
	
	for(int i=0; i<8; ++i)
	{
		origZs[i] = -cubePoints[i][3];

		// TODO: Research homogenized coordinate space clipping
		if(origZs[i] > -nearZ)
		{
			cubePoints[i][3] = nearZ;
			cubePoints[i][2] = 2; // Put it definetely behind the near plane so it gets clipped
		}

		cubePoints[i].Homogenize();
	}
}

bool RenderAnalyzer::PointsWithinBox(const Vector<4>* points, int numPoints) const
{
	// First, frustrum cull the composition by its bounding box
	int dim;
	for(dim=0; dim<3; ++dim)
	{
		int curPoint;
		for(curPoint=0; curPoint<numPoints; ++curPoint)
		{
			if(points[curPoint][dim] > -1) break;
		}
		if(curPoint == 8) break; // Everything's outside the clipping plane, yay
	
		for(curPoint=0; curPoint<numPoints; ++curPoint)
		{
			if(points[curPoint][dim] < 1) break;
		}
		if(curPoint == 8) break; // Everything's outside the clipping plane, yay
	}
	
	if(dim != 3)
	{
		return false; // Culled, no render here... no render children either
	}

	return true;
}


float RenderAnalyzer::GetMaxScreenSpaceArea(const Vector<4> cubePoints[8]) const
{
	Vector<2> minP(1,1);
	Vector<2> maxP(-1,-1);
	for(int i=0; i<8; ++i)
	{
		const Vector<4>& curCubePoint = cubePoints[i];

		if(curCubePoint[0] < minP[0]) minP[0] = curCubePoint[0];
		if(curCubePoint[0] > maxP[0]) maxP[0] = curCubePoint[0];

		if(curCubePoint[1] < minP[1]) minP[1] = curCubePoint[1];
		if(curCubePoint[1] > maxP[1]) maxP[1] = curCubePoint[1];
	}

	Vector<2> diagnol(maxP - minP);

	return diagnol[0] * diagnol[1];
}

float RenderAnalyzer::GetAlphaForArea(float area, float areaThreshold)
{
	const float AREA_END_FADE_PERCENTAGE = 1.6f;
	float endFadeArea = AREA_END_FADE_PERCENTAGE * areaThreshold;
	ASSERT(endFadeArea > areaThreshold);
	if(area < endFadeArea)
	{
		return (area - areaThreshold) / (endFadeArea - areaThreshold);
	}
	else
	{
		return 1.0f;
	}
}

bool RenderAnalyzer::Analyze(SymbolInstance* curInstance, const Matrix<4,4>& modelView, const BoundingBoxProperties& boxProps, const CustomParams& inParams, CustomParams& outParams, int layer)
{
	Matrix<4,4> modelScreen(m_projection*modelView);
	
	float origZs[8];
	Vector<4> cubePoints[8];
	MakeScreenCubePoints(modelScreen, cubePoints, origZs, m_near);

	if(!boxProps.IsInside() && !PointsWithinBox(cubePoints, 8))
	{
		return false;
	}
	
	// Okay, now let's try to figure out what the maximum amount of screen space this object will take up
	float area = GetMaxScreenSpaceArea(cubePoints);

	float alpha;
	if(curInstance->m_sym->GetAreaApproximateThreshold() == 0)
	{
		alpha = 1.0f; // Don't approximate, don't fade.
	}
	else
	{
		alpha = GetAlphaForArea(area, curInstance->m_sym->GetAreaApproximateThreshold());
	}

	if(curInstance->m_sym->GetLSProductionsArray().size() == 0 || area < curInstance->m_sym->GetAreaApproximateThreshold())
	{
		// Just draw the approximation, it's area is too small to recurse
		if(curInstance->m_sym->GetApproxRenderFunc() != 0)
		{
			curInstance->m_sym->GetApproxRenderFunc()(&m_frame, &modelView, inParams.alpha, layer, false);
		}

		return false;
	}
	else 
	{
		outParams.alpha = inParams.alpha;

		// Render this object (fading away) even though we're still recursing
		if(alpha > 0 && alpha < 1)
		{
			if(curInstance->m_prod->GetFM() == Production::FM_FADEOUT
				|| curInstance->m_prod->GetFM() == Production::FM_FADEINOUT)
			{
				// Submit a transparent drawing
				if(curInstance->m_sym->GetApproxRenderFunc() != 0)
				{
					curInstance->m_sym->GetApproxRenderFunc()(&m_frame, &modelView, inParams.alpha * (1-alpha), layer, false);
				}
			}
			if(curInstance->m_prod->GetFM() == Production::FM_FADEIN)
			{
				if(curInstance->m_sym->GetApproxRenderFunc() != 0)
				{
					curInstance->m_sym->GetApproxRenderFunc()(&m_frame, &modelView, inParams.alpha, layer, false);
				}
			}

			if(curInstance->m_prod->GetFM() == Production::FM_FADEIN
				|| curInstance->m_prod->GetFM() == Production::FM_FADEINOUT)
			{
				outParams.alpha = inParams.alpha * alpha;
			}
		}		

		return true;
	}
}

