#include "WorldInstance.h"

#include <R3Utils.h>

#include <BoundedQueue.h>
#include <ThreadPrimitives.h>

#include <Utils.h>
#include <algorithm>

#include <boost/bind.hpp>

#include <iostream>

SymbolInstancePool::BasePool::BasePool(int poolSize)
: m_poolSize(poolSize)
{
	m_poolMemory = static_cast<SymbolInstance*>(operator new(poolSize * sizeof(SymbolInstance)));
	m_curPosition = m_poolMemory;
}

SymbolInstancePool::BasePool::~BasePool()
{
	operator delete(m_poolMemory);
}


SymbolInstance* SymbolInstancePool::BasePool::MakeInstance()
{
	Mutex::ScopedLock lock(m_monitorLock);

	int slotsLeft = m_poolSize - (m_curPosition - m_poolMemory);
	if(slotsLeft < 1) return 0;

	SymbolInstance* retVal = new(m_curPosition) SymbolInstance();
	++m_curPosition;

	return retVal;
}

SymbolInstance* SymbolInstancePool::BasePool::MakeInstanceArray(int numInstances)
{
	Mutex::ScopedLock lock(m_monitorLock);

	int slotsLeft = m_poolSize - (m_curPosition - m_poolMemory);
	if(slotsLeft < numInstances) return 0;

	SymbolInstance* retVal = new(m_curPosition) SymbolInstance[numInstances];
	m_curPosition += numInstances;

	return retVal;
}

const Production::ChildDescriptor* SymbolInstance::GetMyChildDescriptorFromParent()
{
	ASSERT(m_parent != 0);
	ASSERT(m_parent->m_prod != 0);
	for(unsigned i=0; i<m_parent->m_prod->GetChildren().size(); ++i)
	{
		if(&m_parent->m_children[i] == this)
		{
			return m_parent->m_prod->GetChildren()[i];
		}
	}
	
	ASSERT(false);
	return 0;
}

BoundingBoxProperties::BoundingBoxProperties(SymbolInstance* curInstance, const Matrix<4,4>& modelView)
{
	Vector<4> toCenter((modelView.GetColumn(0)+modelView.GetColumn(1)+modelView.GetColumn(2))/2);
	m_avgModelPos = modelView.GetColumn(3) + toCenter;
	
	m_radSquared = toCenter.Trim().LengthSquared(); 
	if(m_avgModelPos.Trim().LengthSquared() < m_radSquared*1.3)
	{
		m_isInside = true;
	}
	else
	{
		m_isInside = false;
	}
}

WorldInstance::WorldInstance(const WorldDefinition& worldDef, int poolSize, Symbol* startSym)
: m_worldDef(worldDef)
{
	// if start sym is 0, we need to find a random symbol
	//if(startSym == 0) startSym = m_worldDef.GetRandomSymbol();
	WorldDefinition::SymbolSet::const_iterator first = m_worldDef.GetSymbols().begin();

	if(startSym == 0) startSym = *first;

	ASSERT(m_worldDef.Contains(startSym));

	// Search through the productions and find out which has the most children
	int mostSymbols = 0;
	for(WorldDefinition::ProductionSet::const_iterator iter = m_worldDef.GetProductions().begin();
		iter != m_worldDef.GetProductions().end(); ++iter)
	{
		const Production& curProd = **iter;

		if(static_cast<int>(curProd.GetChildren().size()) > mostSymbols)
		{
			mostSymbols = curProd.GetChildren().size();
		}
	}

	for(int i=0; i<NUM_POOLS; ++i)
	{
		m_instancePools[i] = new SymbolInstancePool(poolSize, 1 + (MAX_UP_INSTANCES+1)*mostSymbols);
	}
	m_curPool = 0;

	// Make an instance of startSym and use that as our initial pool
	SymbolInstance* startInst = GetCurPool().MakeInstance(true);
	startInst->m_sym = startSym;
	startInst->m_prod = ChooseProductionForNewSymbolInstance(startSym);

	m_refInstance = startInst;
}

WorldInstance::~WorldInstance()
{
	for(int i=0; i<NUM_POOLS; ++i)
	{
		delete m_instancePools[i];
	}
}

void WorldInstance::UseNextPool()
{
	m_curPool = (m_curPool + 1) % NUM_POOLS;
	m_instancePools[m_curPool]->Reset();
}


int WorldInstance::ChooseChildIndexForParent(Production* prod, Symbol* sym, const Vector<4>& curPos)
{
	int winnerIndex = -1;
	Vector<4> winnerPos;
	float winnerScore=0;
			
	int i=0;
	for(Production::ChildList::const_iterator iter = prod->GetChildren().begin();
		iter != prod->GetChildren().end(); ++iter, ++i)
	{
		if((*iter)->sym != sym) continue;
				
		Vector<4> myRelPos = (*iter)->transform * curPos;
		Vector<3> fromCenter = (myRelPos - Vector<4>(0.5,0.5,0.5,1)).Trim();

		// Find the largest component from the center
		float myScore = fromCenter.LengthSquared();
				
		// The winner is the vector who's maximum component is smallest
		if(winnerIndex == -1 || myScore < winnerScore)
		{
			winnerIndex = i;
			winnerPos = myRelPos;
			winnerScore = myScore;
		}
	}
	ASSERT(winnerIndex != -1);

	return winnerIndex;
}


Production* WorldInstance::ChooseProductionForNewSymbolInstance(Symbol* sym)
{
	if(!sym->GetLSProductions().empty())
	{
		const Symbol::ProductionArray& prodArray = sym->GetLSProductionsArray();
		return prodArray[rand() % prodArray.size()];
	}
	else
	{
		return 0;
	}
}

Matrix<4,4> WorldInstance::SetNewRefInstance(SymbolInstance* nextInstance)
{
	ASSERT(GetCurPool().Contains(nextInstance));

	SymbolInstance* oldRef = m_refInstance;

	m_refInstance = nextInstance;

	return FindRelationMatrix(oldRef, m_refInstance);
}

Matrix<4,4> WorldInstance::FindRelationMatrix(SymbolInstance* source, SymbolInstance* dest)
{
	ASSERT(GetCurPool().Contains(source));
	ASSERT(GetCurPool().Contains(dest));
	
	// Find the shortest path from the source to the destination, and remember the transforms
	// along the way
	Matrix<4,4> sourceMatrix(MakeIdentityMatrix());
	Matrix<4,4> destMatrix;
	SymbolInstance* sourceAncestor = source;
	while(true)
	{
		ASSERT(GetCurPool().Contains(sourceAncestor));
		
		SymbolInstance* destAncestor = dest;
		destMatrix = MakeIdentityMatrix();
		while(true)
		{
			ASSERT(GetCurPool().Contains(destAncestor));
			if(destAncestor == sourceAncestor) break;
			if(destAncestor->m_parent == 0) break;
			
			ASSERT(GetCurPool().Contains(destAncestor->m_parent));
			destMatrix = destMatrix * destAncestor->GetMyChildDescriptorFromParent()->invTransform;
			destAncestor = destAncestor->m_parent;
		}

		if(destAncestor == sourceAncestor) break;

		ASSERTMSG(sourceAncestor->m_parent != 0, "Should have found a parent by now");
		sourceMatrix = sourceMatrix * sourceAncestor->GetMyChildDescriptorFromParent()->invTransform;
		sourceAncestor = sourceAncestor->m_parent;
	}
	
	return destMatrix * sourceMatrix.Inverse();
}





